"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import sys

from .cli import main

sys.exit(main())



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import pytest

from cotainr.util import answer_is_yes


class TestAnswerYes:
    @pytest.mark.parametrize("input_text", ["no", "No", "nO", "NO"])
    def test_answering_no(self, input_text, factory_mock_input, monkeypatch):
        monkeypatch.setattr("builtins.input", factory_mock_input(input_text))
        assert not answer_is_yes("some_answer_prompt")

    @pytest.mark.parametrize(
        "input_text", ["yes", "Yes", "yEs", "yeS", "YEs", "YeS", "yES", "YES"]
    )
    def test_answering_yes(self, input_text, factory_mock_input, monkeypatch):
        monkeypatch.setattr("builtins.input", factory_mock_input(input_text))
        assert answer_is_yes("some_answer_prompt")

    @pytest.mark.parametrize(
        "input_text",
        [
            "i",
            "!",
            " ",
            "O!",
            "Anbsolutely not",
            "haha",
            "I don't know",
            "yse",
            "You know it",
            "n",
            "N",
            "nope",
            "NO!",
            "Not gonna happen",
            "y",
            "Y",
        ],
    )
    def test_not_recognized(self, capsys, input_text, factory_mock_input, monkeypatch):
        monkeypatch.setattr("builtins.input", factory_mock_input(input_text))
        assert not answer_is_yes("some_answer_prompt", max_attempts=2)
        assert (
            f'You answered "{input_text}". Please answer yes or no.\n>>> '
            in capsys.readouterr().out
        )

    @pytest.mark.parametrize("max_attempts", [1, 2, 10, 1000])
    def test_too_many_attempts(
        self, max_attempts, factory_mock_input_sequence, monkeypatch, caplog
    ):
        input_sequence = ["Why do I keep answering this question?"] * max_attempts + [
            "yes"
        ]
        monkeypatch.setattr(
            "builtins.input", factory_mock_input_sequence(input_sequence)
        )
        assert not answer_is_yes("some_answer_prompt", max_attempts=max_attempts)
        assert (
            f'You provided an invalid answer {max_attempts} times in a row. Now assuming you meant "no".'
            in caplog.text
        )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import io

import pytest

from cotainr.tracing import StreamWriteProxy


class TestConstructor:
    def test_attributes(self):
        stream = io.StringIO()
        stream_write_proxy = StreamWriteProxy(stream=stream)
        assert stream_write_proxy._stream is stream
        assert stream_write_proxy.true_stream_write == stream.write


class TestWrite:
    def test_write_when_stream_write_changed(self):
        stream = io.StringIO()
        stream_write_proxy = StreamWriteProxy(stream=stream)
        stream.write = lambda s: ""
        stream.write("some text to disappear ")
        stream_write_proxy.write("some text to keep")
        assert stream.getvalue() == "some text to keep"

    def test_write_when_stream_write_unchanged(self):
        stream = io.StringIO()
        stream_write_proxy = StreamWriteProxy(stream=stream)
        stream.write("some text to keep - ")
        stream_write_proxy.write("some text to keep")
        assert stream.getvalue() == "some text to keep - some text to keep"


class TestGetAttr:
    def test_failing_attribute_lookup_delegation(self):
        stream = io.StringIO("some text 6021")
        stream_write_proxy = StreamWriteProxy(stream=stream)
        with pytest.raises(
            AttributeError,
            match=r"StringIO' object has no attribute 'some_nonsense_attribute_6021'$",
        ):
            _ = stream_write_proxy.some_nonsense_attribute_6021

    def test_successful_attribute_lookup_delegation(self):
        stream = io.StringIO("some text 6021")
        stream_write_proxy = StreamWriteProxy(stream=stream)
        assert "getvalue" not in dir(stream_write_proxy)
        assert stream_write_proxy.getvalue() == "some text 6021"



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import logging
import platform
import re
import subprocess
import urllib.error

import pytest

from cotainr.container import SingularitySandbox
from cotainr.pack import CondaInstall
from cotainr.tracing import LogSettings

from ..container.data import data_cached_ubuntu_sif
from ..container.patches import patch_disable_singularity_sandbox_subprocess_runner
from .patches import (
    patch_disable_conda_install_bootstrap_conda,
    patch_disable_conda_install_download_miniforge_installer,
)
from .stubs import StubEmptyLicensePopen, StubShowLicensePopen


class TestConstructor:
    def test_attributes(
        self,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
        assert conda_install.sandbox == sandbox
        assert conda_install.prefix == "/opt/cotainr/conda"
        assert conda_install.license_accepted
        assert conda_install.log_dispatcher is None
        assert conda_install._verbosity == 0

    def test_beforehand_license_acceptance(
        self,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
        capsys,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            CondaInstall(sandbox=sandbox, license_accepted=True)

        # Check that the message about accepting Miniforge license on
        # beforehand is shown
        (
            _sandbox_create_cmd,
            _sandbox_uname_cmd,
            miniforge_license_accept_cmd,
            _conda_bootstrap_cmd,
            _conda_bootstrap_clean_cmd,
        ) = capsys.readouterr().out.strip().split("\n")
        assert miniforge_license_accept_cmd == (
            "You have accepted the Miniforge installer license via the command line option "
            "'--accept-licenses'."
        )

    def test_cleanup(
        self,
        capsys,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)

            # Check that installer was removed after use
            assert not (
                conda_install.sandbox.sandbox_dir / "conda_installer.sh"
            ).exists()

        (
            _sandbox_create_cmd,
            _sandbox_uname_cmd,
            _miniforge_license_accept_cmd,
            bootstrap_cmd,
            clean_cmd,
            _empty_string,
        ) = capsys.readouterr().out.split("\n")

        # Check that installer (mock) was created in the first place
        assert bootstrap_cmd.startswith("PATCH: Bootstrapped Conda using ")
        assert bootstrap_cmd.endswith("conda_installer.sh")

        # Check that the conda clean (mock) command ran
        assert clean_cmd.startswith("PATCH: Ran command in sandbox:")
        assert "'conda', 'clean', '-y', '-a'" in clean_cmd

    def test_setup_log_dispatcher(
        self,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        log_settings = LogSettings(verbosity=1)
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(
                sandbox=sandbox, license_accepted=True, log_settings=log_settings
            )
        assert conda_install._verbosity == 1
        assert conda_install.log_dispatcher.verbosity == log_settings.verbosity
        assert conda_install.log_dispatcher.log_file_path == log_settings.log_file_path
        assert conda_install.log_dispatcher.no_color == log_settings.no_color
        assert conda_install.log_dispatcher.map_log_level is CondaInstall._map_log_level
        assert conda_install.log_dispatcher.logger_stdout.name == "CondaInstall.out"
        assert conda_install.log_dispatcher.logger_stderr.name == "CondaInstall.err"


@pytest.mark.conda_integration
@pytest.mark.singularity_integration
class TestAddEnvironment:
    def test_env_creation(self, tmp_path, data_cached_ubuntu_sif):
        conda_env_path = tmp_path / "conda_env.yml"
        prefix = "/opt/cotainr/conda"
        conda_env_path.write_text(
            "channels:\n  - conda-forge\ndependencies:\n  - python"
        )
        conda_env_name = "some_env_name_6021"
        with SingularitySandbox(base_image=data_cached_ubuntu_sif) as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            conda_install.add_environment(path=conda_env_path, name=conda_env_name)
            process = sandbox.run_command_in_container(cmd="conda info -e")
            assert re.search(
                # We expect to find a line in the list of environments similar to:
                # some_env_name_6021         /opt/cotainr/conda/envs/some_env_name_6021
                rf"^{conda_env_name}(\s)+{prefix}/envs/{conda_env_name}$",
                process.stdout,
                flags=re.MULTILINE,
            )

    def test_other_conda_channels_than_condaforge(
        self, tmp_path, data_cached_ubuntu_sif
    ):
        prefix = "/opt/cotainr/conda"
        conda_env_path = tmp_path / "conda_env.yml"
        conda_env_path.write_text(
            "channels:\n  - bioconda\ndependencies:\n  - samtools"
        )
        conda_env_name = "some_bioconda_env_6021"
        with SingularitySandbox(base_image=data_cached_ubuntu_sif) as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            conda_install.add_environment(path=conda_env_path, name=conda_env_name)
            process = sandbox.run_command_in_container(cmd="conda info -e")
            assert re.search(
                # We expect to find a line in the list of environments similar to:
                # some_env_name_6021         /opt/cotainr/conda/envs/some_env_name_6021
                rf"^{conda_env_name}(\s)+{prefix}/envs/{conda_env_name}$",
                process.stdout,
                flags=re.MULTILINE,
            )


@pytest.mark.conda_integration
@pytest.mark.singularity_integration
class TestCleanupUnusedFiles:
    def test_all_unneeded_removed(self, data_cached_ubuntu_sif):
        with SingularitySandbox(base_image=data_cached_ubuntu_sif) as sandbox:
            CondaInstall(sandbox=sandbox, license_accepted=True)
            process = sandbox.run_command_in_container(cmd="conda clean -d -a")
            clean_msg = (
                "There are no unused tarball(s) to remove.\n"
                "There are no index cache(s) to remove.\n"
                "There are no unused package(s) to remove.\n"
                "There are no tempfile(s) to remove.\n"
                "There are no logfile(s) to remove."
            )
            assert process.stdout.strip() == clean_msg


@pytest.mark.conda_integration
@pytest.mark.singularity_integration
class Test_BootstrapConda:
    def test_correct_conda_installer_bootstrap(self, data_cached_ubuntu_sif):
        with SingularitySandbox(base_image=data_cached_ubuntu_sif) as sandbox:
            prefix = "opt/cotainr/conda"
            conda_install_dir = sandbox.sandbox_dir / prefix
            assert not conda_install_dir.exists()
            CondaInstall(sandbox=sandbox, license_accepted=True)

            # Check that conda has been installed
            assert conda_install_dir.exists()

            # Check that we are using our installed conda
            process = sandbox.run_command_in_container(cmd="conda info --base")
            assert process.stdout.strip() == "/" + prefix

            # Check that the installed conda is up-to-date
            process = sandbox.run_command_in_container(
                cmd="conda update -n base -d conda"
            )
            assert "# All requested packages already installed." in process.stdout


class Test_CheckCondaBootstrapIntegrity:
    def test_bail_on_interfering_conda_installs(
        self,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            with pytest.raises(RuntimeError) as exc_info:
                # Abuse that the output of the mocked subprocess runner in the
                # sandbox is a line about the command run
                # - not the actual output of "conda info --base"
                conda_install._check_conda_bootstrap_integrity()

        exc_msg = str(exc_info.value)
        assert exc_msg.startswith(
            "Multiple Conda installs interfere. "
            "We risk destroying the Conda install in "
            "PATCH: Ran command in sandbox: "
        )
        assert exc_msg.endswith("Aborting!")
        assert "'conda', 'info', '--base'" in exc_msg


class Test_DisplayMessage:
    @pytest.mark.parametrize("level", [logging.DEBUG, logging.INFO])
    def test_log_to_stdout(
        self,
        level,
        caplog,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        log_settings = LogSettings(verbosity=3)
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(
                sandbox=sandbox, license_accepted=True, log_settings=log_settings
            )
            conda_install._display_message(msg="test_6021", log_level=level)

        assert caplog.records[-1].levelno == level
        assert caplog.records[-1].name == "CondaInstall.out"
        assert caplog.records[-1].msg == "test_6021"

    @pytest.mark.parametrize(
        "level", [logging.WARNING, logging.ERROR, logging.CRITICAL]
    )
    def test_log_to_stderr(
        self,
        level,
        caplog,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        log_settings = LogSettings(verbosity=0)
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(
                sandbox=sandbox, license_accepted=True, log_settings=log_settings
            )
            conda_install._display_message(msg="test_6021", log_level=level)

        assert caplog.records[-1].levelno == level
        assert caplog.records[-1].name == "CondaInstall.err"
        assert caplog.records[-1].msg == "test_6021"

    def test_no_log_dispatcher(
        self,
        capsys,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            conda_install._display_message(msg="test_6021", log_level=logging.INFO)

        stdout_lines = capsys.readouterr().out.rstrip("\n").split("\n")
        assert stdout_lines[-1] == "test_6021"

    def test_no_log_level(
        self,
        capsys,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        log_settings = LogSettings(verbosity=1)
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(
                sandbox=sandbox, license_accepted=True, log_settings=log_settings
            )
            conda_install._display_message(msg="test_6021")

        stdout_lines = capsys.readouterr().out.rstrip("\n").split("\n")
        assert stdout_lines[-1] == "test_6021"


class Test_DisplayMiniforgeLicenseForAcceptance:
    @pytest.mark.parametrize("answer", ["yes", "Yes", "YES", "YeS"])
    def test_accepting_license(
        self,
        answer,
        factory_mock_input,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
        capsys,
        monkeypatch,
    ):
        monkeypatch.setattr(subprocess, "Popen", StubShowLicensePopen)
        monkeypatch.setattr("builtins.input", factory_mock_input(answer))
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox)

        stdout_lines = capsys.readouterr().out.replace(">>> ", "").strip().split("\n")
        assert "You have accepted the Miniforge installer license." in stdout_lines
        assert conda_install.license_accepted

    @pytest.mark.parametrize("answer", ["n", "N", ""])
    def test_not_accepting_license(
        self,
        answer,
        factory_mock_input,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
        capsys,
        monkeypatch,
    ):
        monkeypatch.setattr(subprocess, "Popen", StubShowLicensePopen)
        monkeypatch.setattr("builtins.input", factory_mock_input(answer))
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            with pytest.raises(SystemExit):
                CondaInstall(sandbox=sandbox)

        stdout_lines = capsys.readouterr().out.replace(">>> ", "").strip().split("\n")
        assert (
            "You have not accepted the Miniforge installer license. Aborting!"
        ) in stdout_lines

    def test_installer_not_showing_license(
        self,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
        monkeypatch,
    ):
        monkeypatch.setattr(subprocess, "Popen", StubEmptyLicensePopen)
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            with pytest.raises(
                RuntimeError,
                match="^No license seems to be displayed by the Miniforge installer.$",
            ):
                CondaInstall(sandbox=sandbox)

    def test_license_interaction_handling(
        self,
        factory_mock_input,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
        capsys,
        monkeypatch,
    ):
        monkeypatch.setattr(subprocess, "Popen", StubShowLicensePopen)
        monkeypatch.setattr("builtins.input", factory_mock_input("yes"))
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            CondaInstall(sandbox=sandbox)

        stdout = capsys.readouterr().out.strip()
        stdout_lines = stdout.split("\n")

        # Check that we send the "ENTER" and kill the process.
        assert "StubShowLicensePopen received: input='\\n'." in stdout_lines
        assert "StubShowLicensePopen killed." in stdout_lines

        # Check that the prompt for "ENTER" is not displayed to the user
        assert "Please, press ENTER to continue\n>>> " not in stdout

        # Check that the license is displayed to the user
        assert "STUB:\n\n\nThis is the license terms..." in stdout

    @pytest.mark.conda_integration
    def test_miniforge_still_showing_license(
        self,
        factory_mock_input,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_singularity_sandbox_subprocess_runner,
        capsys,
        monkeypatch,
    ):
        monkeypatch.setattr("builtins.input", factory_mock_input("yes"))
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            sandbox.architecture = platform.machine()
            CondaInstall(sandbox=sandbox)

        stdout = capsys.readouterr().out.strip()

        # Check that various lines that we expect to always be part of
        # https://github.com/conda-forge/miniforge/blob/main/LICENSE are still
        # shown to the user when extracting and showing the Miniforge license
        # terms from the installer - just to have an idea that the license is
        # still being shown correctly to the user
        license_parts = [
            "Miniforge installer comes",
            "installing miniforge and is deleted",
            "bootstrapping executable uses",
            "micromamba",
            "conda-forge",
            "All rights reserved.",
        ]
        for license_part in license_parts:
            assert license_part in stdout


class Test_DownloadMiniforgeInstaller:
    def test_installer_download_success(
        self,
        patch_urllib_urlopen_as_bytes_stream,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            # Just needs some supported architecture
            sandbox.architecture = "x86_64"
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            conda_installer_path = (
                conda_install.sandbox.sandbox_dir / "conda_installer_download"
            )
            conda_install._download_miniforge_installer(
                installer_path=conda_installer_path
            )
            conda_installer_strings = [
                b"PATCH: Bytes returned by urlopen for url=",
                b"'https://github.com/conda-forge/miniforge/releases/latest/download/Miniforge3-",
            ]
            conda_installer_bytes_string = conda_installer_path.read_bytes()
            for installer_string in conda_installer_strings:
                assert installer_string in conda_installer_bytes_string

    @pytest.mark.conda_integration  # technically not a test that depends on Conda - but a very slow one...
    def test_installer_download_fail(
        self,
        patch_urllib_urlopen_force_fail,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            # Just needs some supported architecture
            sandbox.architecture = "x86_64"
            with pytest.raises(
                urllib.error.URLError, match="PATCH: urlopen error forced for url="
            ):
                CondaInstall(sandbox=sandbox)

    def test_unknown_architecture(
        self, patch_disable_singularity_sandbox_subprocess_runner
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            sandbox.architecture = None
            with pytest.raises(
                RuntimeError,
                match=(
                    r".*"
                    r"Cotainr.*"
                    r"CondaInstall.*"
                    r"which indicates that it is not running in a container sandbox context"
                ),
            ):
                CondaInstall(sandbox=sandbox, license_accepted=True)


class Test_GetInstallScript:
    @pytest.mark.parametrize("arch", ["arm64", "aarch64"])
    def test_arm_success(self, arch):
        assert CondaInstall._get_install_script(arch) == "Miniforge3-Linux-aarch64.sh"

    def test_x86_success(self):
        assert (
            CondaInstall._get_install_script("x86_64") == "Miniforge3-Linux-x86_64.sh"
        )

    @pytest.mark.parametrize(
        "arch",
        [
            "test",
            "arms64",
            "aarchs64",
            "None",
            "Weird",
            "WINDOWS",
            "icecream",
            "86_64",
            "pc",
        ],
    )
    def test_unknown_arch_error(self, arch):
        with pytest.raises(
            ValueError,
            match=(
                r".*"
                r"Cotainr.*"
                r"CondaInstall.*"
                r"supports.*"
                r"x86_64.*"
                r"arm64.*"
                r"aarch64.*"
                rf"{arch}.*"
            ),
        ):
            CondaInstall._get_install_script(arch)


class Test_CondaVerbosityArg:
    @pytest.mark.parametrize(
        ["verbosity", "verbosity_arg"],
        [
            (-1, " -q"),
            (0, " -q"),
            (1, ""),
            (2, " -v"),
            (3, " -vv"),
            (4, " -vv"),
            (5, " -vvv"),
        ],
    )
    def test_correct_mapping_of_verbosity(
        self,
        verbosity,
        verbosity_arg,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            conda_install._verbosity = verbosity
            assert conda_install._conda_verbosity_arg == verbosity_arg


class Test_LoggingFilters:
    def test_correctly_ordered_list_of_filters(
        self,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)

        filter_names = [
            filter_.__class__.__name__ for filter_ in conda_install._logging_filters
        ]
        assert filter_names == [
            "StripANSIEscapeCodes",
            "OnlyFinalProgressbarFilter",
            "NoEmptyLinesFilter",
        ]

    def test_strip_ANSI_codes_filter(
        self,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            filter_ = conda_install._logging_filters[0]
            assert filter_.__class__.__name__ == "StripANSIEscapeCodes"

        rec = logging.LogRecord(
            name="test",
            level=logging.INFO,
            pathname="test_path",
            lineno=0,
            msg=(
                "\x1b[38;5;8msome_\x1b[38;5;3m\x1b[38;5;1m\033[38;5;160mlong\x1b[0m_message_"
                "\033[Awith\x1b[A_a_lot_of_AN\x1b[KSI_codes_\x1b[B6021"
            ),
            args=None,
            exc_info=None,
        )
        filter_.filter(rec)
        assert rec.msg == "some_long_message_with_a_lot_of_ANSI_codes_6021"

    @pytest.mark.parametrize(["msg", "keep"], [("", False), ("some_msg_6021", True)])
    def test_no_empty_lines_filter(
        self,
        msg,
        keep,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            filter_ = conda_install._logging_filters[2]
            assert filter_.__class__.__name__ == "NoEmptyLinesFilter"

        rec = logging.LogRecord(
            name="test",
            level=logging.INFO,
            pathname="test_path",
            lineno=0,
            msg=msg,
            args=None,
            exc_info=None,
        )
        assert filter_.filter(rec) == keep

    @pytest.mark.parametrize(
        ["msg", "keep"],
        [
            (
                "CondaInstall.out:-: openssl-3.1.3        | 2.5 MB    |            |   0%",
                False,
            ),
            (
                "CondaInstall.out:-: jsonpatch-1.33       | 17 KB     | #########4 |  94%",
                False,
            ),
            (
                "CondaInstall.out:-: libnsl-2.0.0         | 32 KB     | ####9      |  49%",
                False,
            ),
            (
                "CondaInstall.out:-: zstandard-0.21.0     | 395 KB    | 4          |   4%",
                False,
            ),
            (
                "CondaInstall.out:-: python-3.11.5        | 29.4 MB   | ########## | 100%",
                True,
            ),
        ],
    )
    def test_only_final_progressbar_filter(
        self,
        msg,
        keep,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_singularity_sandbox_subprocess_runner,
    ):
        with SingularitySandbox(base_image="my_base_image_6021") as sandbox:
            conda_install = CondaInstall(sandbox=sandbox, license_accepted=True)
            filter_ = conda_install._logging_filters[1]
            assert filter_.__class__.__name__ == "OnlyFinalProgressbarFilter"

        rec = logging.LogRecord(
            name="test",
            level=logging.INFO,
            pathname="test_path",
            lineno=0,
            msg=msg,
            args=None,
            exc_info=None,
        )
        assert filter_.filter(rec) == keep


class Test_MapLogLevel:
    @pytest.mark.parametrize(
        ["msg", "log_level"],
        [
            ("DEBUG", logging.DEBUG),
            ("VERBOSE", logging.DEBUG),
            ("TRACE", logging.DEBUG),
            ("DEBUG:6021", logging.DEBUG),
            ("VERBOSE 6021", logging.DEBUG),
            ("TRACEING my day away", logging.DEBUG),
            ("INFO", logging.INFO),
            ("INFORMATION", logging.INFO),
            ("WARNING", logging.WARNING),
            ("WARNING-log", logging.WARNING),
            ("ERROR", logging.ERROR),
            ("ERROR_MSG", logging.ERROR),
            ("CRITICAL", logging.CRITICAL),
            ("CRITICAL/mission", logging.CRITICAL),
            ("unknown", logging.INFO),
            ("debug something", logging.INFO),
            (
                "some messages containing DEBUG, VERBOSE, TRACE, WARNING, ERROR, and CRITICAL",
                logging.INFO,
            ),
        ],
    )
    def test_correct_log_level_mapping(self, msg, log_level):
        assert CondaInstall._map_log_level(msg=msg) == log_level



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import functools
import io
import logging
import platform
import subprocess
import sys

import pytest

from cotainr.tracing import LogDispatcher, LogSettings
from cotainr.util import _print_and_capture_stream, stream_subprocess


class TestStreamSubprocess:
    def test_completed_process(self):
        proc_args = [
            sys.executable,
            "-c",
            "import platform; print(platform.python_version())",
        ]
        process = stream_subprocess(args=proc_args)
        assert process.args == proc_args
        assert process.returncode == 0
        assert process.stdout.strip() == platform.python_version()

    def test_check_returncode(self):
        cmd_exit = [sys.executable, "-c", "import sys; sys.exit(1)"]
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            stream_subprocess(args=cmd_exit)

        assert exc_info.value.returncode == 1
        assert exc_info.value.cmd == cmd_exit

    def test_kwargs_passing(self):
        env = {"COTAINR_TEST_ENV_VAR": "6021"}
        process = stream_subprocess(
            args=[
                sys.executable,
                "-c",
                "import os; print(os.environ['COTAINR_TEST_ENV_VAR'])",
            ],
            env=env,
        )
        assert process.stdout.strip() == env["COTAINR_TEST_ENV_VAR"]

    def test_logging_stdout(self, caplog):
        log_level = logging.INFO
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: log_level,
            log_settings=LogSettings(verbosity=1, log_file_path=None, no_color=False),
        )
        caplog.clear()  # Clear any logs from LogDispatcher initialization
        stdout_text = """
        Text on line 1

        More text later!
        """

        stream_subprocess(
            args=[
                sys.executable,
                "-c",
                f"import os; os.write(1, {stdout_text.encode()})",
            ],
            log_dispatcher=log_dispatcher,
        )

        stdout_lines = stdout_text.splitlines(keepends=True)
        assert len(stdout_lines) == len(caplog.records)
        for line, rec in zip(stdout_lines, caplog.records):
            assert rec.levelno == log_level
            assert rec.name == "test_dispatcher_6021.out"
            assert rec.msg == line

    def test_logging_stderr(self, caplog):
        log_level = logging.ERROR
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: log_level,
            log_settings=LogSettings(verbosity=0, log_file_path=None, no_color=False),
        )
        caplog.clear()  # Clear any logs from LogDispatcher initialization
        stderr_text = """
        An error!
        More breakage...
        """

        stream_subprocess(
            args=[
                sys.executable,
                "-c",
                f"import os; os.write(2, {stderr_text.encode()})",
            ],
            log_dispatcher=log_dispatcher,
        )

        stderr_lines = stderr_text.splitlines(keepends=True)
        assert len(stderr_lines) == len(caplog.records)
        for line, rec in zip(stderr_lines, caplog.records):
            assert rec.levelno == log_level
            assert rec.name == "test_dispatcher_6021.err"
            assert rec.msg == line

    def test_stdout(self):
        stdout_text = """
        Text on line 1

        More text later!
        """
        process = stream_subprocess(
            args=[
                sys.executable,
                "-c",
                f"import os; os.write(1, {stdout_text.encode()})",
            ]
        )
        assert process.stdout == stdout_text

    def test_stderr(self):
        stderr_text = """
        An error!
        More breakage...
        """
        process = stream_subprocess(
            args=[
                sys.executable,
                "-c",
                f"import os; os.write(2, {stderr_text.encode()})",
            ]
        )
        assert process.stderr == stderr_text


class Test_PrintAndCaptureStream:
    def test_print_stdout_roundtrip(self, capsys):
        stdout_lines = ["Test line 1", "Test line 2", "Test line 3"]
        input_stream = io.StringIO("\n".join(stdout_lines))
        print_dispatch = functools.partial(print, end="", file=sys.stdout)
        _print_and_capture_stream(
            stream_handle=input_stream, print_dispatch=print_dispatch
        )
        captured_io = capsys.readouterr()
        assert captured_io.out.split("\n") == stdout_lines

    def test_print_and_capture_equality(self):
        input_text = "Test line 1\nTest line 2"
        input_stream = io.StringIO(input_text)
        output_stream = io.StringIO()
        print_dispatch = functools.partial(print, end="", file=output_stream)
        captured_stream = _print_and_capture_stream(
            stream_handle=input_stream, print_dispatch=print_dispatch
        )
        assert output_stream.getvalue() == "".join(captured_stream)



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import shlex

import pytest

from cotainr import __version__ as _cotainr_version
from cotainr.cli import CotainrCLI

from .container.data import data_cached_ubuntu_sif


@pytest.mark.endtoend
@pytest.mark.parametrize(
    "umask",
    [
        # umask  file permissions  directory permissions
        0o007,  # 660 (rw-rw----)   770 (rwxrwx---)  # umask used on LUMI
        0o022,  # 644 (rw-r--r--)   755 (rwxr-xr-x)  # default umask in most cases
    ],
)
def test_conda_env_build(
    umask,
    context_set_umask,
    data_cached_ubuntu_sif,
    singularity_exec,
    singularity_inspect,
    tmp_path,
):
    with context_set_umask(umask):  # default umask on LUMI
        """Test build command when including simple conda environment."""
        build_container_path = tmp_path / "conda_container.sif"
        conda_env_path = tmp_path / "conda_env.yml"
        conda_env_path.write_text(
            "channels:\n  - conda-forge\ndependencies:\n  - python"
        )

        CotainrCLI(
            args=shlex.split(
                f"build {build_container_path} "
                f"--base-image={data_cached_ubuntu_sif} "
                f"--conda-env={conda_env_path} "
                "--accept-licenses"
            )
        ).subcommand.execute()

        container_metadata = singularity_inspect(build_container_path).stdout.split(
            "\n"
        )
        assert f"cotainr.version: {_cotainr_version}" in container_metadata
        assert "cotainr.url: https://github.com/DeiC-HPC/cotainr" in container_metadata
        assert any(
            datum.startswith("cotainr.command: ") for datum in container_metadata
        )

        container_python_process = singularity_exec(
            f"{build_container_path} python3 -c 'import sys; print(sys.executable)'"
        )
        for conda_identifier in ["conda", "/envs/"]:
            assert conda_identifier in container_python_process.stdout



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import shutil
from subprocess import CompletedProcess

import pytest

import cotainr.container


@pytest.fixture
def patch_disable_singularity_sandbox_subprocess_runner(monkeypatch):
    """
    Disable SingularitySandbox._subprocess_runner(...).

    The subprocess runner is replaced by a method that prints and returns a
    message about the command that would have been run.
    """

    def mock_subprocess_runner(self, *, args, **kwargs):
        msg = f"PATCH: Ran command in sandbox: {args}"
        print(msg)
        return CompletedProcess(args, returncode=0, stdout=msg)

    monkeypatch.setattr(
        cotainr.container.SingularitySandbox,
        "_subprocess_runner",
        mock_subprocess_runner,
    )


@pytest.fixture
def patch_fake_singularity_sandbox_env_folder(monkeypatch):
    """
    Fake the creation of the .singularity.d/env/ folder.
    Additionally, fake creation of the environment file 92-cotainr-env.sh

    Normally, the folder is created by SingularitySandbox.__enter__() when it
    runs Singularity to create the sandbox, but when the call to singularity
    has been patched, this fixture may be used to create the folder anyway.
    When singularity is patched, the environment file can no longer be created
    inside the container with correct file permissions, so here it is created outside.
    """

    def mock_enter(self):
        # Call "true" __enter__ for setup
        ret_val = self._non_mocked_context_enter()

        # Create fake environment folder
        singularity_env_folder = self.sandbox_dir / ".singularity.d/env"
        singularity_env_folder.mkdir(parents=True, exist_ok=True)

        return ret_val

    def mock_create_file(self, f):
        # Create file outside the container
        env_file = f or self.sandbox_dir / ".singularity.d/env/92-cotainr-env.sh"
        env_file.touch(exist_ok=True)

    monkeypatch.setattr(
        cotainr.container.SingularitySandbox,
        "_create_file",
        mock_create_file,
    )

    monkeypatch.setattr(
        cotainr.container.SingularitySandbox,
        "_non_mocked_context_enter",
        cotainr.container.SingularitySandbox.__enter__,
        raising=False,
    )

    monkeypatch.setattr(cotainr.container.SingularitySandbox, "__enter__", mock_enter)


@pytest.fixture
def patch_save_singularity_sandbox_context(monkeypatch):
    """
    Store the SingularitySandbox context for inspection in tests.

    The content on `SingularitySandbox.sandbox_dir` is copied to
    `saved_sandbox_dir` before being cleaned up.
    """
    saved_sandbox_dir_name = "saved_sandbox_dir"

    def mock_exit(self, exc_type, exc_value, traceback):
        # Copy content of _tmp_dir
        shutil.copytree(self.sandbox_dir, self._origin / saved_sandbox_dir_name)

        # Call "true" __exit__ for cleanup
        return self._non_mocked_context_exit(exc_type, exc_value, traceback)

    monkeypatch.setattr(
        cotainr.container.SingularitySandbox,
        "_non_mocked_context_exit",
        cotainr.container.SingularitySandbox.__exit__,
        raising=False,
    )
    monkeypatch.setattr(cotainr.container.SingularitySandbox, "__exit__", mock_exit)

    return saved_sandbox_dir_name


@pytest.fixture
def patch_disable_add_metadata(monkeypatch):
    """
    Disable SingularitySandbox.add_metadata().
    """
    monkeypatch.setattr(
        cotainr.container.SingularitySandbox, "add_metadata", lambda _: None
    )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import builtins
import contextlib
import io
import logging
import os
from pathlib import Path
import shlex
import subprocess
import sys
import urllib.error
import urllib.request

import pytest


@pytest.fixture
def argparse_options_line():
    """
    Return the help text line for optional arguments in argparse.

    Apparently this line changed from "optional arguments" to "options" in Python 3.10.
    """
    py_ver = sys.version_info
    if py_ver.major > 3 or (py_ver.major == 3 and py_ver.minor >= 10):
        return "options:\n"
    else:
        return "optional arguments:\n"


@pytest.fixture
def context_importerror(monkeypatch):
    """
    Force an ImportError when importing the a specified module.

    The `mock_import_context` contextmanager forces an `ImportError` to be
    raised when trying to import the `module_name` module within the context.
    Imports of all other modules are unaffected.
    """

    @contextlib.contextmanager
    def mock_import_context(module_name):
        builtins_import = builtins.__import__

        def mock_import(name, globals, locals, fromlist, level):
            if name == module_name:
                raise ImportError(f"PATCH: ImportError forced for {module_name=}")
            else:
                return builtins_import(name, globals, locals, fromlist, level)

        with monkeypatch.context() as m:
            m.setattr("builtins.__import__", mock_import)
            yield

    return mock_import_context


@pytest.fixture(autouse=True)
def context_reload_logging():
    """
    Reset the internal state of the logging module on test teardown.

    Needed in tests of logging functionality where the tests end up affecting
    the internal state of the logging module.

    The current implementation is based on:
    https://til.tafkas.net/posts/-resetting-python-logging-before-running-tests/

    It used to be implemented as `importlib.reload(logging)`. However, it turns
    out that this simpler approach does not work for some tests where multiple
    test cases use the caplog fixture. For those cases, only the first test
    case would actually capture the log messages - the others would be empty.
    See this GH PR comment for an in-depth discussion of the issue:
    https://github.com/DeiC-HPC/cotainr/pull/154#discussion_r2166428044
    """
    yield
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    loggers.append(logging.getLogger())
    for logger in loggers:
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)
            handler.close()
        logger.setLevel(logging.NOTSET)
        logger.propagate = True
        logger.manager.loggerDict = {}


@pytest.fixture
def context_set_umask():
    """Return a context manager providing a context with the specified umask."""

    @contextlib.contextmanager
    def set_umask(umask):
        current_umask = None
        try:
            current_umask = os.umask(umask)
            yield
        finally:
            if current_umask is not None:
                os.umask(current_umask)

    return set_umask


@pytest.fixture
def data_log_level_names_mapping():
    """
    A mapping from log levels to their names.

    From Python 3.11 this mapping is also available as
    logging.getLevelNamesMapping().
    """
    level_names_mapping = {
        level: logging.getLevelName(level)
        for level in [
            logging.CRITICAL,
            logging.ERROR,
            logging.WARNING,
            logging.INFO,
            logging.DEBUG,
        ]
    }

    return level_names_mapping


@pytest.fixture
def factory_mock_input():
    """
    Create mock of the builtins `input` function that returns a fixed "input".

    Returns a factory for creating mocked versions of the builtin `input`
    function to be used with the `monkeypatch` fixture to replace
    `builtins.input` with a function that prints the prompt (its argument, if
    provided) and returns a "fixed user input", provided as argument to the
    factory.
    """

    def create_mock_input(fixed_user_input=None):
        def mock_input(prompt):
            print(prompt, end="")
            return fixed_user_input

        return mock_input

    return create_mock_input


@pytest.fixture
def factory_mock_input_sequence():
    """
    Create mock of the builtins `input` function that returns a sequence of "inputs".

    Returns a factory for creating mocked versions of the builtin `input`
    function to be used with the `monkeypatch` fixture to replace
    `builtins.input` with a function that prints the prompt (its argument, if
    provided) and returns the "next user input" from a sequence, provided as
    argument to the factory.

    If more inputs are requested than provided in the sequence, a
    `StopIteration` exception is raised.
    """

    def create_mock_input(user_input_sequence=(None,)):
        inputs = iter(user_input_sequence)

        def mock_input(prompt):
            print(prompt, end="")
            return next(inputs)

        return mock_input

    return create_mock_input


@pytest.fixture
def patch_urllib_urlopen_as_bytes_stream(monkeypatch):
    """
    Disable urllib.request.urlopen(...).

    The `urlopen` contextmanager is replaced by a mock that returns a bytes
    message about the URL content that would have been opened.
    """

    @contextlib.contextmanager
    def mock_urlopen(url, *args, **kwargs):
        yield io.BytesIO(f"PATCH: Bytes returned by urlopen for {url=}".encode())

    monkeypatch.setattr(urllib.request, "urlopen", mock_urlopen)


@pytest.fixture
def patch_urllib_urlopen_force_fail(monkeypatch):
    """
    Force urllib.request.urlopen(...) to raise URLError.

    The `urlopen` contextmanager is replaced by a mock that always raises an
    `urllib.error.URLError` when entering the contextmanager.
    """

    @contextlib.contextmanager
    def mock_urlopen(url, *args, **kwargs):
        raise urllib.error.URLError(f"PATCH: urlopen error forced for {url=}")

    monkeypatch.setattr(urllib.request, "urlopen", mock_urlopen)


@pytest.fixture(autouse=True)
def safedir(tmp_path):
    """
    Force test to be run in a temporary directory.

    A lot of the functionality in cotainr manipulates directories. In order to
    provide some protection against messing up the current working directory by
    running a test, we force all tests to run from a temporary directory by
    making this an "autouse" fixture.
    """
    origin = Path().resolve()
    safe_dir = tmp_path / "safe_dir"
    safe_dir.mkdir()
    os.chdir(safe_dir)
    yield
    os.chdir(origin)


@pytest.fixture
def singularity_exec():
    """
    Provide a function wrapping a "singularity exec" call.

    The function returned by this fixture provides a shorthand to run
    "singularity exec {cmd}" as a python subprocess.
    """

    def _singularity_exec(cmd):
        singularity_process = subprocess.run(
            ["singularity", "exec", *shlex.split(cmd)],
            capture_output=True,
            check=True,
            text=True,
        )

        return singularity_process

    return _singularity_exec


@pytest.fixture
def singularity_inspect():
    """Provide a function wrapping a "singularity inspect" call."""
    return lambda container_path: subprocess.run(
        ["singularity", "inspect", container_path],
        capture_output=True,
        check=True,
        text=True,
    )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import logging

import pytest

from cotainr.tracing import ColoredOutputFormatter


class TestClassAttributes:
    def test_log_level_fg_colors(self):
        assert hasattr(ColoredOutputFormatter, "log_level_fg_colors")
        assert isinstance(ColoredOutputFormatter.log_level_fg_colors, dict)
        for log_level in [
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
            logging.CRITICAL,
        ]:
            assert log_level in ColoredOutputFormatter.log_level_fg_colors

    def test_reset_color(self):
        assert hasattr(ColoredOutputFormatter, "reset_color")
        assert ColoredOutputFormatter.reset_color == "\x1b[0m"


class TestFormat:
    @pytest.mark.parametrize(
        ["level", "msg", "color_msg"],
        [
            (logging.DEBUG, "test_debug", "\x1b[38;5;8mtest_debug\x1b[0m"),
            (logging.INFO, "test_info", "test_info"),
            (logging.WARNING, "test_warning", "\x1b[38;5;3mtest_warning\x1b[0m"),
            (logging.ERROR, "test_error", "\x1b[38;5;1mtest_error\x1b[0m"),
            (logging.CRITICAL, "test_critical", "\x1b[38;5;160mtest_critical\x1b[0m"),
        ],
    )
    def test_correct_formatting(self, level, msg, color_msg):
        formatter = ColoredOutputFormatter()
        rec = logging.LogRecord(
            name="test",
            level=level,
            pathname="test_path",
            lineno=0,
            msg=msg,
            args=None,
            exc_info=None,
        )
        assert formatter.format(rec) == color_msg

    def test_original_record_unchanged(self):
        formatter = ColoredOutputFormatter()
        rec = logging.LogRecord(
            name="test",
            level=logging.ERROR,
            pathname="test_path",
            lineno=0,
            msg="test_msg",
            args=None,
            exc_info=None,
        )
        msg = formatter.format(rec)
        assert msg is not rec.msg
        assert rec.msg == "test_msg"



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import datetime
import inspect
import locale
from subprocess import CompletedProcess

import pytest

from .prepare_release import (
    _check_release_version_format,
    _format_release_notes_date,
    create_docs_switcher,
    create_release_notes,
    format_release_version_and_date,
)


class Test__check_release_version_format:
    def test_acceptable_YYYY_0M_MICRO_format(self):
        assert _check_release_version_format(release_version="2025.03.1")

    def test_unacceptable_YYYY_0M_MICRO_format(self):
        with pytest.raises(ValueError):
            _check_release_version_format(release_version="2025.04.0")

    @pytest.mark.parametrize(
        "valid_version_number", ["2023.12.0", "2025.1.0", "2025.2.0"]
    )
    def test_correct_YYYY_MM_MICRO_format(self, valid_version_number):
        assert _check_release_version_format(release_version=valid_version_number)

    @pytest.mark.parametrize(
        "invalid_version_number", ["1900.2.0", "2025.32.1", "2025.2.01"]
    )
    def test_wrong_format(self, invalid_version_number):
        with pytest.raises(ValueError):
            _check_release_version_format(release_version=invalid_version_number)

    def test_not_raising_error(self):
        assert not _check_release_version_format(
            release_version="1999.1.0", raise_error=False
        )


class Test_create_docs_switcher:
    def test_correctly_formatted_switcher(self, monkeypatch):
        def mock_subprocess_run(*args, **kwargs):
            return CompletedProcess(
                args="", returncode=0, stdout="2022.1.0\n2022.02.0\n2022.11.0\n"
            )

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        expected_switcher = inspect.cleandoc(  # Remove leading whitespace
            """
            [
              {
                "name": "dev",
                "version": "latest",
                "url": "https://cotainr.readthedocs.io/en/latest/"
              },
              {
                "name": "2025.1.0 (stable)",
                "version": "stable",
                "url": "https://cotainr.readthedocs.io/en/stable/"
              },
              {
                "name": "2022.11.0",
                "version": "2022.11.0",
                "url": "https://cotainr.readthedocs.io/en/2022.11.0/"
              },
              {
                "name": "2022.02.0",
                "version": "2022.02.0",
                "url": "https://cotainr.readthedocs.io/en/2022.02.0/"
              },
              {
                "name": "2022.1.0",
                "version": "2022.1.0",
                "url": "https://cotainr.readthedocs.io/en/2022.1.0/"
              }
            ]
            """
        )
        switcher = create_docs_switcher(formatted_release_version="2025.1.0")
        assert switcher == expected_switcher

    def test_invalid_version_number(self):
        with pytest.raises(ValueError):
            create_docs_switcher(formatted_release_version="hello 6021!")


class Test_create_release_notes:
    def test_correctly_formatted_output(self):
        expected_release_notes = (
            inspect.cleandoc(  # Remove leading whitespace
                """
            # 2025.1.0

            **Released on January 1st, 2025**

            [Documentation for this release](https://cotainr.readthedocs.org/en/2025.1.0/)

            __short summary__

            ## New features

            - __new feature (externals) description__, __link to GitHub PR__

            ## Bug fixes

            - __bug fix description__, __link to GitHub PR__

            ## Maintenance updates

            - __maintenance update (internals) description__, __link to GitHub PR__
            """
            )
            + "\n"  # Newline at the end of the file
        )
        release_notes = create_release_notes(
            formatted_release_version="2025.1.0",
            formatted_release_date="January 1st, 2025",
        )
        assert release_notes == expected_release_notes

    def test_invalid_version_number(self):
        with pytest.raises(ValueError):
            create_release_notes(
                formatted_release_version="hello 6021!",
                formatted_release_date="invalid date",
            )


class Test_format_release_version_and_date:
    def test_correctly_formatted_date(self, monkeypatch):
        monkeypatch.setattr("cotainr.__version__", "2025.03.0")
        _, formatted_release_date = format_release_version_and_date(
            release_date="2025-04-01"
        )
        assert formatted_release_date == "April 1st, 2025"

    @pytest.mark.parametrize(
        ["current_release_version", "release_date", "new_formatted_release_version"],
        [
            (
                "2025.1.1",
                "2025-01-15",
                "2025.1.1",
            ),  # would be automatically set to 2025.1.1 by hatch-vcs
            (
                "2025.1.2",
                "2025-01-16",
                "2025.1.2",
            ),  # would be automatically set to 2025.1.1 by hatch-vcs
            ("2025.1.2", "2025-03-17", "2025.3.0"),
            (
                "2025.12.3",
                "2025-12-22",
                "2025.12.3",
            ),  # would be automatically set to 2025.1.1 by hatch-vcs
            ("2025.12.3", "2026-01-23", "2026.1.0"),
            ("2025.1.0", "2026-11-20", "2026.11.0"),
        ],
    )
    def test_correctly_formatted_version(
        self,
        current_release_version,
        release_date,
        new_formatted_release_version,
        monkeypatch,
    ):
        monkeypatch.setattr("cotainr.__version__", current_release_version)
        formatted_release_version, _ = format_release_version_and_date(
            release_date=release_date
        )

        assert formatted_release_version == new_formatted_release_version

    @pytest.mark.parametrize(
        ["current_release_version", "release_date"],
        [
            ("2025.1.0", "2024-01-01"),
            ("2025.2.0", "2025-01-31"),
            ("2025.2.0", "2024-01-10"),
        ],
    )
    def test_invalid_release_date(
        self, current_release_version, release_date, monkeypatch
    ):
        monkeypatch.setattr("cotainr.__version__", current_release_version)
        with pytest.raises(
            ValueError,
            match=(
                f"New release date {release_date} is before the current version "
                f"{current_release_version}."
            ),
        ):
            format_release_version_and_date(release_date=release_date)

    def test_release_today(self, monkeypatch):
        class MockedDatetime(datetime.datetime):
            @classmethod
            def today(cls):
                return cls(2025, 10, 1)

        monkeypatch.setattr(datetime, "datetime", MockedDatetime)
        monkeypatch.setattr("cotainr.__version__", "2025.1.0")
        formatted_release_version, formatted_release_date = (
            format_release_version_and_date()
        )
        assert formatted_release_version == "2025.10.0"
        assert formatted_release_date == "October 1st, 2025"

    def test_unknown_current_version(self, monkeypatch):
        monkeypatch.setattr("cotainr.__version__", "<unknown version>")
        with pytest.raises(
            ValueError,
            match="Current release version <unknown version> is not in the expected format.",
        ):
            format_release_version_and_date(release_date="2025-01-01")


class Test_format_release_notes_date:
    @pytest.mark.parametrize(
        ["iso_input", "formatted_date"],
        [
            ("2021-01-01", "January 1st, 2021"),
            ("2021-01-02", "January 2nd, 2021"),
            ("2021-01-03", "January 3rd, 2021"),
            ("2022-02-04", "February 4th, 2022"),
            ("2022-02-05", "February 5th, 2022"),
            ("2022-02-06", "February 6th, 2022"),
            ("2023-03-07", "March 7th, 2023"),
            ("2023-03-08", "March 8th, 2023"),
            ("2023-03-09", "March 9th, 2023"),
            ("2023-03-10", "March 10th, 2023"),
            ("2024-04-11", "April 11th, 2024"),
            ("2024-04-12", "April 12th, 2024"),
            ("2024-04-13", "April 13th, 2024"),
            ("2025-05-14", "May 14th, 2025"),
            ("2025-05-15", "May 15th, 2025"),
            ("2025-05-16", "May 16th, 2025"),
            ("2026-06-17", "June 17th, 2026"),
            ("2026-06-18", "June 18th, 2026"),
            ("2026-06-19", "June 19th, 2026"),
            ("2027-07-20", "July 20th, 2027"),
            ("2027-07-21", "July 21st, 2027"),
            ("2027-07-22", "July 22nd, 2027"),
            ("2028-08-23", "August 23rd, 2028"),
            ("2028-08-24", "August 24th, 2028"),
            ("2028-08-25", "August 25th, 2028"),
            ("2029-09-26", "September 26th, 2029"),
            ("2029-09-27", "September 27th, 2029"),
            ("2030-10-28", "October 28th, 2030"),
            ("2031-11-29", "November 29th, 2031"),
            ("2032-12-30", "December 30th, 2032"),
            ("2032-12-31", "December 31st, 2032"),
        ],
    )
    def test_date_formatting_default_locale(self, iso_input, formatted_date):
        assert (
            _format_release_notes_date(datetime.date.fromisoformat(iso_input))
            == formatted_date
        )

    def test_error_on_dk_locale(self, monkeypatch):
        monkeypatch.setattr(locale, "getlocale", lambda: ("da_DK", "UTF-8"))
        with pytest.raises(RuntimeError, match="Your locale is set to"):
            _format_release_notes_date(datetime.date.fromisoformat("2021-01-01"))



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import contextlib


class AlwaysCompareFalse:
    """A stub that returns False for all Python "rich comparisons"."""

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False

    def __eq__(self, other):
        return False

    def __ne__(self, other):
        return False

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __repr__(self):
        return "AlwaysCompareFalseStub"


class FixedNumberOfSpinsEvent:
    """
    An event that fixes the number of times the message is spun in a MessageSpinner.
    """

    def __init__(self, *, spins):
        assert isinstance(spins, int)
        assert spins >= 0
        self.remaining_spins = spins

    def is_set(self):
        if self.remaining_spins == 0:
            return True
        else:
            self.remaining_spins -= 1
            return False

    def set(self):
        pass


@contextlib.contextmanager
def RaiseOnEnterContext():
    raise NotImplementedError("Entered context")
    yield



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

from pathlib import Path
import subprocess
import sys

import pytest

import cotainr

from .cli.patches import patch_disable_main


class Test__main__:
    def test_programmatic_import_run(self, patch_disable_main):
        with pytest.raises(SystemExit, match="^PATCH: Ran cotainr.cli main function$"):
            from cotainr import __main__

    def test_cli_m_flag(self):
        env = {
            # cotainr is not importable anymore when the safedir fixture changes
            # the working directory during the test invocation.
            "PYTHONPATH": f"{Path(cotainr.__file__).parent.parent}"
        }
        process = subprocess.run(
            [sys.executable, "-m", "cotainr"],
            capture_output=True,
            check=True,
            text=True,
            env=env,
        )
        assert process.stdout.startswith("usage: cotainr [-h]")



r"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

This module implements utility functions.

Functions
---------
answer_is_yes()
    Ask user for confirmation ("yes") of `input_text`.
get_systems()
    Get a dictionary of predefined systems, defined in systems.json
stream_subprocess(\*, args, \*\*kwargs)
    Run a the command described by `args` while streaming stdout and stderr.

Attributes
----------
systems_file
    The path to the systems.json file (if present).
"""

from concurrent.futures import ThreadPoolExecutor
import functools
import json
import logging
from pathlib import Path
import subprocess
import sys

logger = logging.getLogger(__name__)
systems_file = (Path(__file__) / "../../systems.json").resolve()


def answer_is_yes(input_text, max_attempts=1000):
    """
    Ask user for confirmation ("yes") of `input_text`.

    Parameters
    ----------
    input_text : str
        The prompt to be printed for verification by the user.
    max_attempts : int, optional
        The maximum number of attempts to get a valid answer from the user before giving up. The default is 1000.

    Returns
    -------
    answer_is_yes : bool
        The indicator of whether or not the answer is yes.
    """
    answer_prompt = input_text + " [yes/no]\n>>> "
    for _ in range(max_attempts):
        answer = input(answer_prompt)
        if answer.lower() == "yes":
            return True
        elif answer.lower() == "no":
            return False
        else:
            answer_prompt = f'You answered "{answer}". Please answer yes or no.\n>>> '
    else:
        logger.error(
            'You provided an invalid answer %d times in a row. Now assuming you meant "no".',
            max_attempts,
        )
        return False


def get_systems():
    """
    Get a dictionary of predefined systems, defined in the systems.json file.

    Returns
    -------
    systems : Dict
        A dictionary of predefined systems

    Raises
    ------
    :class:`NameError`
        If some required arguments are missing in the systems.json file.
    """
    if systems_file.is_file():
        systems = json.loads(systems_file.read_text())
        for system in systems:
            if "base-image" not in systems[system]:
                raise NameError(
                    f"Error in systems.json: {system} missing argument base-image"
                )
        return systems
    else:
        return {}


def stream_subprocess(*, args, log_dispatcher=None, **kwargs):
    """
    Run a the command described by `args` while streaming stdout and stderr.

    The command described by `args` is run in a subprocess with that
    subprocess' stdout and stderr streamed to the main process. Each line in
    the subprocess' output is streamed separately to the main process. Extra
    `kwargs` are passed to `Popen` when opening the subprocess.

    Parameters
    ----------
    args : list or str
        Program arguments. See the docstring for :class:`subprocess.Popen` for
        details.
    log_dispatcher : :class:`~cotainr.tracing.LogDispatcher`, optional
        The log dispatcher which subprocess stdout/stderr messages are
        forwarded to (the default is None which implies that subprocess
        messages are forwarded directly to stdout/stderr).

    Returns
    -------
    completed_process : :class:`subprocess.CompletedProcess`
        Information about the completed subprocess.

    Raises
    ------
    :class:`subprocess.CalledProcessError`
        If the subprocess returned a non-zero status code.

    Notes
    -----
    The way we handle stdout and stderr messages in separate threads introduces
    a race condition in cases where subprocesses write to stdout and stderr at
    the same time. All messages are always guaranteed to be streamed to the
    console, but they might arrive in the wrong order. We accept this "feature"
    as is.
    """
    with subprocess.Popen(
        args,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=1,
        **kwargs,
    ) as process:
        with ThreadPoolExecutor(
            max_workers=2,
            thread_name_prefix=f"cotainr_stream_subprocess_thread_for_{args}",
        ) as executor:
            # (Attempt to) pass the process stdout and stderr to the terminal in real
            # time while also storing it for later inspection.
            stdout_future = executor.submit(
                _print_and_capture_stream,
                stream_handle=process.stdout,
                print_dispatch=(
                    log_dispatcher.log_to_stdout
                    if log_dispatcher is not None
                    else functools.partial(print, end="", file=sys.stdout)
                ),
            )
            stderr_future = executor.submit(
                _print_and_capture_stream,
                stream_handle=process.stderr,
                print_dispatch=(
                    log_dispatcher.log_to_stderr
                    if log_dispatcher is not None
                    else functools.partial(print, end="", file=sys.stderr)
                ),
            )
            captured_stdout = stdout_future.result()
            captured_stderr = stderr_future.result()

    completed_process = subprocess.CompletedProcess(
        process.args,
        process.returncode,
        stdout="".join(captured_stdout),
        stderr="".join(captured_stderr),
    )

    completed_process.check_returncode()

    return completed_process


def _print_and_capture_stream(*, stream_handle, print_dispatch):
    """
    Print a text stream while also storing it.

    Parameters
    ----------
    stream_handle : :py:class:`io.TextIOWrapper`
        The text stream to print and capture.
    print_dispatch : Callable
        The callable to use for printing.

    Returns
    -------
    captured_stream : list of str
        The lines captured from the stream.
    """
    captured_stream = []
    for line in stream_handle:
        print_dispatch(line)
        captured_stream.append(line)

    return captured_stream


def _flush_stdin_buffer():
    """
    Discard queued data on stdin file descriptor.

    TCIOFLUSH selects both the input queue and output queue to be discarded.

    https://stackoverflow.com/questions/2520893/how-to-flush-the-input-stream
    """
    from io import UnsupportedOperation

    try:
        sys.stdin.fileno()
    except UnsupportedOperation:
        # stdin is a pseudofile (e.g. Pytest without --no-capture)
        return

    # Python Standard library, Linux/Unix/OSX
    from termios import TCIOFLUSH, tcflush

    tcflush(sys.stdin, TCIOFLUSH)



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import pytest

import cotainr.util

from .patches import (
    patch_empty_system,
    patch_system_with_actual_file,
    patch_system_with_badly_formatted_file,
    patch_system_with_non_existing_file,
)


class TestGetSystems:
    def test_non_existing(self, patch_system_with_non_existing_file):
        systems = cotainr.util.get_systems()

        assert isinstance(systems, dict)
        assert not systems

    def test_non_empty(self, patch_empty_system):
        systems = cotainr.util.get_systems()

        assert isinstance(systems, dict)
        assert not systems

    def test_actual_file(self, patch_system_with_actual_file):
        systems = cotainr.util.get_systems()

        assert isinstance(systems, dict)
        assert len(systems) == 2
        assert list(systems.keys())[0].startswith("some")
        assert list(systems.keys())[1].startswith("another")

    def test_badly_formatted_file(self, patch_system_with_badly_formatted_file):
        with pytest.raises(
            NameError,
            match="Error in systems.json: some_system_6021 missing argument base-image",
        ):
            cotainr.util.get_systems()



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import concurrent.futures
import sys

import pytest

from cotainr.tracing import ConsoleSpinner, StreamWriteProxy, console_lock

from .patches import patch_fix_number_of_message_spins
from .stubs import RaiseOnEnterContext


class TestConstructor:
    def test_attributes(self):
        console_spinner = ConsoleSpinner()
        assert isinstance(console_spinner._stdout_proxy, StreamWriteProxy)
        assert console_spinner._stdout_proxy._stream is sys.stdout
        assert isinstance(console_spinner._stderr_proxy, StreamWriteProxy)
        assert console_spinner._stderr_proxy._stream is sys.stderr
        assert console_spinner._true_input_func is input
        assert console_spinner._spinning_msg is None
        assert console_spinner._in_nested_context_count == 0


class TestContext:
    def test_correctly_handling_input(self, factory_mock_input, capsys, monkeypatch):
        monkeypatch.setattr("builtins.input", factory_mock_input("some_input_6021"))
        with ConsoleSpinner():
            input("please_input_some_text_6021")

        stdout = capsys.readouterr().out
        assert stdout == "please_input_some_text_6021"

    def test_correctly_setting_nested_context_count(self):
        console_spinner = ConsoleSpinner()
        assert not console_lock.locked()
        with ConsoleSpinner() as level_1_spinner:
            assert console_lock.locked()
            assert level_1_spinner._in_nested_context_count == 0
            with console_spinner as outer_spinner:
                assert console_lock.locked()
                assert level_1_spinner._in_nested_context_count == 0
                assert outer_spinner._in_nested_context_count == 1
                with console_spinner:
                    assert console_lock.locked()
                    assert level_1_spinner._in_nested_context_count == 0
                    assert outer_spinner._in_nested_context_count == 2
                    with ConsoleSpinner() as level_2_spinner:
                        assert console_lock.locked()
                        assert level_1_spinner._in_nested_context_count == 0
                        assert outer_spinner._in_nested_context_count == 2
                        assert level_2_spinner._in_nested_context_count == 1

                    assert console_lock.locked()
                    assert level_1_spinner._in_nested_context_count == 0
                    assert outer_spinner._in_nested_context_count == 2

                assert console_lock.locked()
                assert level_1_spinner._in_nested_context_count == 0
                assert outer_spinner._in_nested_context_count == 1

            assert console_lock.locked()
            assert level_1_spinner._in_nested_context_count == 0

        assert not console_lock.locked()

    @pytest.mark.parametrize("spins", [1])
    def test_correctly_spinning_stdout_msg(
        self, spins, capsys, patch_fix_number_of_message_spins
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        with ConsoleSpinner():
            sys.stdout.write("test_6021")

        stdout = capsys.readouterr().out
        spin_and_final_line = "\r\x1b[2K⣾ test_6021..." + "\r\x1b[2Ktest_6021\n"
        assert stdout == spin_and_final_line

    @pytest.mark.parametrize("spins", [1])
    def test_correctly_spinning_stderr_msg(
        self, spins, capsys, patch_fix_number_of_message_spins
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        with ConsoleSpinner():
            sys.stderr.write("test_6021")

        stderr = capsys.readouterr().err
        spin_and_final_line = "\r\x1b[2K⣾ test_6021..." + "\r\x1b[2Ktest_6021\n"
        assert stderr == spin_and_final_line

    @pytest.mark.parametrize("spins", [1])
    def test_create_entering_multiple_context_swap(
        self,
        spins,
        capsys,
        monkeypatch,
        factory_mock_input,
        patch_fix_number_of_message_spins,
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        monkeypatch.setattr("builtins.input", factory_mock_input())
        console_spinner_1 = ConsoleSpinner()
        console_spinner_2 = ConsoleSpinner()
        with console_spinner_2:
            sys.stdout.write("test_6021_level_1")
            with console_spinner_1:
                input("level_5_test_input_6021")
                sys.stdout.write("test_6021_level_2")

        stdout = capsys.readouterr().out
        level_1_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_level_1..." + "\r\x1b[2Ktest_6021_level_1\n"
        )
        level_2_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_level_2..." + "\r\x1b[2Ktest_6021_level_2\n"
        )
        assert stdout == (
            level_1_spin_and_final_line
            + "level_5_test_input_6021"
            + level_2_spin_and_final_line
        )

    @pytest.mark.parametrize("spins", [1])
    def test_entering_multiple_context(
        self,
        spins,
        capsys,
        monkeypatch,
        factory_mock_input,
        patch_fix_number_of_message_spins,
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        monkeypatch.setattr("builtins.input", factory_mock_input())
        console_spinner = ConsoleSpinner()
        with ConsoleSpinner():
            sys.stdout.write("test_6021_level_1")
            with console_spinner:
                sys.stdout.write("test_6021_level_2")
                with console_spinner:
                    with ConsoleSpinner():
                        sys.stdout.write("test_6021_level_4")
                        with ConsoleSpinner():
                            input("level_5_test_input_6021")
                            with ConsoleSpinner(), console_spinner:
                                sys.stdout.write("test_6021_level_7")

        stdout = capsys.readouterr().out
        level_1_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_level_1..." + "\r\x1b[2Ktest_6021_level_1\n"
        )
        level_2_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_level_2..." + "\r\x1b[2Ktest_6021_level_2\n"
        )
        level_4_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_level_4..." + "\r\x1b[2Ktest_6021_level_4\n"
        )
        level_7_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_level_7..." + "\r\x1b[2Ktest_6021_level_7\n"
        )
        assert stdout == (
            level_1_spin_and_final_line
            + level_2_spin_and_final_line
            + level_4_spin_and_final_line
            + "level_5_test_input_6021"
            + level_7_spin_and_final_line
        )

    def test_exiting_context_with_spinning_msg(self, capsys):
        with ConsoleSpinner():
            sys.stdout.write("test_6021")

        stdout, stderr = capsys.readouterr()
        assert stdout.endswith("\r\x1b[2Ktest_6021\n")
        assert not stderr

    def test_exiting_context_without_spinning_msg(self, capsys):
        with ConsoleSpinner():
            pass

        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert not stderr

    def test_lock_enter(self):
        console_spinner = ConsoleSpinner()
        console_spinner._lock = RaiseOnEnterContext()
        with pytest.raises(NotImplementedError, match=r"^Entered context$"):
            console_spinner.__enter__()

    def test_lock_exit(self):
        console_spinner = ConsoleSpinner()
        console_spinner._lock = RaiseOnEnterContext()
        with pytest.raises(NotImplementedError, match=r"^Entered context$"):
            console_spinner.__exit__(None, None, None)

    def test_return_self(self):
        with ConsoleSpinner() as console_spinner:
            assert isinstance(console_spinner, ConsoleSpinner)

    @pytest.mark.parametrize("spins", [1])
    def test_switching_from_stdout_to_stderr(
        self, spins, capsys, patch_fix_number_of_message_spins
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        with ConsoleSpinner():
            sys.stdout.write("test_6021_stdout")
            sys.stderr.write("test_6021_stderr")

        stdout, stderr = capsys.readouterr()
        stdout_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_stdout..." + "\r\x1b[2Ktest_6021_stdout\n"
        )
        stderr_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_stderr..." + "\r\x1b[2Ktest_6021_stderr\n"
        )
        assert stdout == stdout_spin_and_final_line
        assert stderr == stderr_spin_and_final_line

    @pytest.mark.parametrize("spins", [1])
    def test_switching_from_stderr_to_stdout(
        self, spins, capsys, patch_fix_number_of_message_spins
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        with ConsoleSpinner():
            sys.stderr.write("test_6021_stderr")
            sys.stdout.write("test_6021_stdout")

        stdout, stderr = capsys.readouterr()
        stdout_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_stdout..." + "\r\x1b[2Ktest_6021_stdout\n"
        )
        stderr_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_stderr..." + "\r\x1b[2Ktest_6021_stderr\n"
        )
        assert stdout == stdout_spin_and_final_line
        assert stderr == stderr_spin_and_final_line

    @pytest.mark.parametrize(
        ["spins", "num_tasks"], [(1, 1), (1, 5), (1, 10), (1, 20), (1, 100), (1, 200)]
    )
    def test_threadpool_console_spinner_race_condition(
        self, spins, num_tasks, capsys, patch_fix_number_of_message_spins
    ):
        # This is a bit of a wacky test for the case where one starts multiple
        # threads with individual ConsoleSpinner objects. Only one
        # ConsoleSpinner object can control the stdout manipulation, so when
        # this "main" context is exited in one thread, the spinner disappears
        # in all other threads as well if they haven't managed to start the
        # spinning message yet. This creates a race condition which we only
        # handle by making sure that the message is indeed still printed,
        # though without a spinner. The number of threads that don't get to
        # have a spinner becomes random which is why we test multiple values
        # for the number of tasks in the hope that at least one of them trigger
        # this race condition.

        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        def spin_msg_func():
            with ConsoleSpinner():
                sys.stdout.write("test_6021")

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=num_tasks, thread_name_prefix="console_spinner_thread"
        ) as executor:
            futures = [executor.submit(spin_msg_func) for _ in range(num_tasks)]

        stdout = capsys.readouterr().out
        stdout_lines = stdout.split("\n")
        spin_and_final_line = "\r\x1b[2K⣾ test_6021..." + "\r\x1b[2Ktest_6021"
        assert all(future.exception() is None for future in futures)
        for line in stdout_lines[:-1]:
            assert line == spin_and_final_line
        assert stdout_lines[-1] == "test_6021" * (num_tasks - (len(stdout_lines) - 1))

    @pytest.mark.parametrize(["spins", "num_tasks"], [(1, 1), (1, 5), (1, 10)])
    def test_threadpool_same_console_spinner(
        self, spins, num_tasks, capsys, patch_fix_number_of_message_spins
    ):
        # Having all these ConsoleSpinnner context wrap each other get very
        # costly in terms of lock acquiring times which is why limit ourself to
        # 10 tasks max in this unit test.

        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        def spin_msg_func():
            with ConsoleSpinner():
                sys.stdout.write("test_6021")

        with ConsoleSpinner():
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=num_tasks, thread_name_prefix="console_spinner_thread"
            ) as executor:
                futures = [executor.submit(spin_msg_func) for _ in range(num_tasks)]

        stdout = capsys.readouterr().out
        spin_and_final_line = "\r\x1b[2K⣾ test_6021..." + "\r\x1b[2Ktest_6021\n"
        assert all(future.exception() is None for future in futures)
        assert stdout == spin_and_final_line * num_tasks

    @pytest.mark.parametrize("spins", [1])
    def test_updating_via_print_function(
        self, spins, capsys, patch_fix_number_of_message_spins
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.

        # print() works by making two writes to the file/stream:
        # 1. The actual message to write
        # 2. The `end` keyword
        with ConsoleSpinner():
            print("test_6021")

        stdout = capsys.readouterr().out
        spin_and_final_line = "\r\x1b[2K⣾ test_6021..." + "\r\x1b[2Ktest_6021\n"
        assert stdout == spin_and_final_line


class Test_UpdateSpinnerMsg:
    def test_lock(self):
        console_spinner = ConsoleSpinner()
        console_spinner._lock = RaiseOnEnterContext()
        with pytest.raises(NotImplementedError, match=r"^Entered context$"):
            console_spinner._update_spinner_msg("", stream=None)

    @pytest.mark.parametrize("spins", [1])
    def test_first_message_correctly_started(
        self, spins, capsys, patch_fix_number_of_message_spins
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        with console_lock:
            console_spinner = ConsoleSpinner()
            console_spinner._update_spinner_msg("test_6021", stream=sys.stdout)
            console_spinner._spinning_msg.stop()

        stdout = capsys.readouterr().out
        spin_and_final_line = "\r\x1b[2K⣾ test_6021..." + "\r\x1b[2Ktest_6021\n"
        assert stdout == spin_and_final_line

    @pytest.mark.parametrize("spins", [1])
    def test_message_correctly_updated(
        self, spins, capsys, patch_fix_number_of_message_spins
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        with console_lock:
            console_spinner = ConsoleSpinner()
            console_spinner._update_spinner_msg("test_6021_line_1", stream=sys.stdout)
            console_spinner._update_spinner_msg("test_6021_line_2", stream=sys.stdout)
            console_spinner._spinning_msg.stop()

        stdout = capsys.readouterr().out
        msg_1_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_line_1..." + "\r\x1b[2Ktest_6021_line_1\n"
        )
        msg_2_spin_and_final_line = (
            "\r\x1b[2K⣾ test_6021_line_2..." + "\r\x1b[2Ktest_6021_line_2\n"
        )
        assert stdout == msg_1_spin_and_final_line + msg_2_spin_and_final_line

    @pytest.mark.parametrize("spins", [1])
    def test_not_updating_on_empty_message(
        self, spins, capsys, patch_fix_number_of_message_spins
    ):
        # The `spins` parameter is passed automagically to the
        # `patch_fix_number_of_message_spins` fixture.
        with console_lock:
            console_spinner = ConsoleSpinner()
            console_spinner._update_spinner_msg("test_6021", stream=sys.stdout)
            console_spinner._update_spinner_msg("", stream=sys.stdout)
            console_spinner._update_spinner_msg("\n", stream=sys.stdout)
            console_spinner._update_spinner_msg("\t", stream=sys.stdout)
            console_spinner._update_spinner_msg("      ", stream=sys.stdout)
            console_spinner._spinning_msg.stop()

        stdout = capsys.readouterr().out
        spin_and_final_line = "\r\x1b[2K⣾ test_6021..." + "\r\x1b[2Ktest_6021\n"
        assert stdout == spin_and_final_line

    def test_updating_message_outside_console_lock(self, capsys):
        console_spinner = ConsoleSpinner()
        console_spinner._update_spinner_msg("test_6021", stream=sys.stdout)
        stdout = capsys.readouterr().out
        assert stdout == "test_6021"


class Test_ThreadSafeInput:
    def test_lock(self):
        console_spinner = ConsoleSpinner()
        console_spinner._lock = RaiseOnEnterContext()
        wrapped_input_func = console_spinner._thread_safe_input(lambda: None)
        with pytest.raises(NotImplementedError, match=r"^Entered context$"):
            wrapped_input_func()

    def test_stop_spinning_msg(self):
        class DummyMsg:
            def stop(self):
                pass

        console_spinner = ConsoleSpinner()
        console_spinner._spinning_msg = DummyMsg()
        with console_spinner:
            console_spinner._thread_safe_input(lambda: None)()
            assert console_spinner._spinning_msg is None

    def test_wrap_input_func(self):
        def my_input_func():
            return "test_6021"

        with ConsoleSpinner() as console_spinner:
            assert console_spinner._thread_safe_input(my_input_func)() == "test_6021"



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import subprocess

import pytest


@pytest.fixture(scope="session")
def data_cached_alpine_sif(tmp_path_factory):
    """A session scope cached SIF image of the latest alpine linux container."""
    singularity_image_path = (
        tmp_path_factory.mktemp("singularity_images") / "alpine_latest.sif"
    )
    subprocess.run(
        args=[
            "singularity",
            "pull",
            f"{singularity_image_path.resolve()}",
            "docker://alpine:latest",
        ],
        capture_output=True,
        check=True,
        text=True,
    )
    assert singularity_image_path.exists()

    return singularity_image_path


@pytest.fixture(scope="session")
def data_cached_ubuntu_sif(tmp_path_factory):
    """A session scope cached SIF image of the latest ubuntu linux container."""
    singularity_image_path = (
        tmp_path_factory.mktemp("singularity_images") / "ubuntu_latest.sif"
    )
    subprocess.run(
        args=[
            "singularity",
            "pull",
            f"{singularity_image_path.resolve()}",
            "docker://ubuntu:latest",
        ],
        capture_output=True,
        check=True,
        text=True,
    )
    assert singularity_image_path.exists()

    return singularity_image_path



r"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

This module implements the command line main command and subcommands.

Classes
-------
CotainrSubcommand(ABC)
    Abstract base class for `CotainrCLI` subcommands.
Build(CotainrSubcommand)
    Build a container. (The "build" subcommand.)
CotainrCLI
    Build Apptainer/Singularity containers for HPC systems in user space. (The
    main CLI command.)
Info(CotainrSubcommand)
    Obtain info about the state of all required dependencies for building a
    container. (The "info" subcommand.)

Functions
---------
main(\*args, \*\*kwargs)
    Main CLI entrypoint.
"""

from abc import ABC, abstractmethod
import argparse
from datetime import datetime
import logging
from pathlib import Path
import platform
import re
import shutil
import subprocess
import sys
import time

from . import __version__ as _cotainr_version
from . import _minimum_dependency_version as _min_dep_ver
from . import container, pack, tracing, util

logger = logging.getLogger(__name__)


class CotainrSubcommand(ABC):
    """Abstract base class for `CotainrCLI` subcommands."""

    @classmethod  # noqa: B027
    def add_arguments(cls, *, parser):
        """
        Add command line arguments to arguments `parser`.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser to add arguments to.
        """
        pass

    @abstractmethod
    def execute(self):
        """Execute the subcommand."""
        pass  # pragma: no cover


class Build(CotainrSubcommand):
    """
    Build a container.

    The "build" subcommand.

    Parameters
    ----------
    image_path : :class:`os.PathLike`
        Path to the built container image.
    base_image : str
        Base image to use for the container which may be any valid
        Apptainer/Singularity <BUILD SPEC>.
    conda_env : :class:`os.PathLike`, optional
        Path to a Conda environment.yml file to install and activate in the
        container. When installing a Conda environment, you must accept the
        Miniforge license terms, as specified during the build process.
    system : str
        Which system/partition you will be running the container on. This sets
        base image and other parameters for a simpler container creation.
        Running the info command will tell you more about the system and what
        is available.
    accept_licenses : bool, default=False
        Accept all license terms (if any) needed for completing the container
        build process.
    verbosity : int, optional
        The verbosity of the output from cotainr: -1 for only CRITICAL, 0 (the
        default) for cotainr INFO and subprocess WARNING, 1 for subprocess
        output as well, 2 for subprocess INFO, 3 for DEBUG, and 4 for TRACE.
    log_to_file : bool
        Create files containing all logging information shown on stdout/stderr.
    no_color : bool
        Do not use colored console output.
    """

    def __init__(
        self,
        *,
        image_path,
        base_image=None,
        conda_env=None,
        system=None,
        accept_licenses=False,
        verbosity=0,
        log_to_file=False,
        no_color=False,
    ):
        """Construct the "build" subcommand."""
        self.log_settings = tracing.LogSettings(
            verbosity=verbosity,
            log_file_path=(
                Path(".").resolve()
                / f"cotainr_build_{datetime.now().isoformat().replace(':', '.')}"
                if log_to_file
                else None
            ),
            no_color=no_color,
        )
        self.image_path = Path(image_path).resolve()
        if self.image_path.exists():
            overwrite_text = (
                f"{self.image_path} already exists. Would you like to overwrite it?"
            )
            if not util.answer_is_yes(overwrite_text):
                logger.critical(
                    "You have chosen not to overwrite %s. Exiting.", self.image_path
                )
                sys.exit(0)

        self.accept_licenses = accept_licenses
        self.base_image = base_image
        systems = util.get_systems()
        if system is not None:
            if system in systems:
                self.system = systems[system]
                self.base_image = self.system["base-image"]
            else:
                raise KeyError("System does not exist")
        if conda_env is not None:
            self.conda_env = Path(conda_env).resolve()
            if not self.conda_env.exists():
                raise FileNotFoundError(
                    f"The provided Conda env file '{self.conda_env}' does not exist."
                )
        else:
            self.conda_env = None

    @classmethod
    def add_arguments(cls, *, parser):
        """Add arguments to the "build" subcommand subparser."""
        parser.add_argument(
            "image_path",
            help=_extract_help_from_docstring(arg="image_path", docstring=cls.__doc__),
            type=Path,
        )
        system_group = parser.add_mutually_exclusive_group(required=True)
        system_group.add_argument(
            "--base-image",
            help=_extract_help_from_docstring(arg="base_image", docstring=cls.__doc__),
        )
        system_group.add_argument(
            "--system",
            help=_extract_help_from_docstring(arg="system", docstring=cls.__doc__),
        )
        parser.add_argument(
            "--conda-env",
            help=_extract_help_from_docstring(arg="conda_env", docstring=cls.__doc__),
            type=Path,
        )
        parser.add_argument(
            "--accept-licenses",
            help=_extract_help_from_docstring(
                arg="accept_licenses", docstring=cls.__doc__
            ),
            action="store_true",
        )
        verbose_quiet_group = parser.add_mutually_exclusive_group()
        verbose_quiet_group.add_argument(
            "--verbose",
            "-v",
            action="count",
            dest="verbosity",
            default=0,
            help=(
                "increase the verbosity of the output from cotainr. "
                "Can be used multiple times: Once for subprocess output, "
                "twice for subprocess INFO, three times for VERBOSE, four times for DEBUG "
                "and five times for TRACE."
            ),
        )
        verbose_quiet_group.add_argument(
            "--quiet",
            "-q",
            action="store_const",
            const=-1,
            dest="verbosity",
            help="do not show any non-CRITICAL output from cotainr",
        )
        parser.add_argument(
            "--log-to-file",
            action="store_true",
            help=_extract_help_from_docstring(arg="log_to_file", docstring=cls.__doc__),
        )
        parser.add_argument(
            "--no-color",
            action="store_true",
            help=_extract_help_from_docstring(arg="no_color", docstring=cls.__doc__),
        )

    def execute(self):
        """Execute the "build" subcommand."""
        t_start_build = time.time()
        with tracing.ConsoleSpinner():
            logger.info("Creating Singularity Sandbox")
            with container.SingularitySandbox(
                base_image=self.base_image, log_settings=self.log_settings
            ) as sandbox:
                if self.conda_env is not None:
                    # Install supplied conda env
                    logger.info("Installing Conda environment: %s", self.conda_env)
                    conda_env_name = "conda_container_env"
                    conda_env_file = sandbox.sandbox_dir / self.conda_env.name
                    shutil.copyfile(self.conda_env, conda_env_file)
                    conda_install = pack.CondaInstall(
                        sandbox=sandbox,
                        license_accepted=self.accept_licenses,
                        log_settings=self.log_settings,
                    )
                    conda_install.add_environment(
                        path=conda_env_file, name=conda_env_name
                    )

                    sandbox.add_to_env(shell_script=f"conda activate {conda_env_name}")

                    # Clean-up unused files
                    logger.info("Cleaning up unused Conda files")
                    conda_install.cleanup_unused_files()
                    logger.info(
                        "Finished installing conda environment: %s", self.conda_env
                    )

                logger.info("Adding metadata to container")
                sandbox.add_metadata()
                logger.info("Building container image")
                sandbox.build_image(path=self.image_path)

            t_end_build = time.time()
            logger.info(
                "Finished building %s in %s",
                self.image_path,
                time.strftime("%H:%M:%S", time.gmtime(t_end_build - t_start_build)),
            )


class Info(CotainrSubcommand):
    """
    Obtain info about the state of all required dependencies for building a container.

    The "info" subcommand.
    """

    def __init__(self):
        """Construct the "info" subcommand."""
        self._checkmark = "\x1b[38;5;2mOK\x1b[0m"  # green OK
        self._nocheckmark = "\x1b[38;5;1mERROR\x1b[0m"  # red ERROR
        self._tabs_width = 4

    def execute(self):
        """Execute the "info" subcommand."""
        print("Dependency report")
        print("-" * 79)
        print(f"\t- {self._check_python_dependency()}".expandtabs(self._tabs_width))
        print(
            f"\t- {self._check_singularity_dependency()}".expandtabs(self._tabs_width)
        )
        print("")
        print("System info")
        print("-" * 79)
        print(self._check_systems())

    def _check_python_dependency(self):
        """
        Check and report on the Python version used.

        Reports the Python version used and whether it meets the minimum
        required version.

        Returns
        -------
        python_check_result : str
            A description of the Python version used.
        """
        ver_tup = platform.python_version_tuple()
        ver_check = self._check_version(
            version=tuple(map(int, ver_tup)), min_version=_min_dep_ver["python"]
        )
        return f"Running python {'.'.join(ver_tup)} {ver_check}"

    def _check_singularity_dependency(self):
        """
        Check and report on any installed singularity provider.

        Reports the provider (apptainer/singularity/unknown), its version, and
        whether it meets the minimum required version if a singularity provider
        is found. Otherwise reports "apptainer/singularity not found".

        Returns
        -------
        singularity_check_result : str
            A description of the singularity provider.

        Notes
        -----
        Assumes that "singularity --version" returns a format like:
          - singularity-ce version 3.11.4-1 (for singularity community edition)
          - apptainer version 1.0.3      (for apptainer)
        """
        try:
            provider, _, version = subprocess.check_output(
                ["singularity", "--version"], text=True
            ).split()
            ver_tup = tuple(
                map(int, re.findall(r"\d+\.\d+\.\d+", version)[0].split("."))
            )
            if provider in _min_dep_ver:
                ver_check = self._check_version(
                    version=ver_tup, min_version=_min_dep_ver[provider]
                )
                singularity_check_result = f"Found {provider} {version} {ver_check}"
            else:
                singularity_check_result = (
                    f"Found unknown singularity provider: {provider} {version}"
                )
        except (subprocess.CalledProcessError, FileNotFoundError):
            singularity_check_result = (
                f"apptainer/singularity not found, {self._nocheckmark}"
            )

        return singularity_check_result

    def _check_systems(self):
        """
        Check and report on any available HPC systems.

        Reports the available systems from the `systems.json` file.

        Returns
        -------
        system_check_result : str
            A description of the available system configurations.
        """
        systems = util.get_systems()
        system_check_report = []
        if systems:
            system_check_report.append("Available system configurations:")
            for system in systems:
                system_check_report.append(f"\t- {system}".expandtabs(self._tabs_width))
        else:
            system_check_report.append("No system configurations available")

        return "\n".join(system_check_report)

    def _check_version(self, *, version, min_version):
        """
        Check and report a version against a minimum required version.

        Checks whether `version` is larger than or equal to `min_version` and
        returns a description of result. The version specifications must be
        (major, minor, patchlevel) tuples of integers.

        Parameters
        ----------
        version : tuple of int
            The version to check against a minimum version.
        min_version : tuple of int
            The minimum version to check against.

        Returns
        -------
        ver_check : str
            A description of the result of the version check.

        Raises
        ------
        TypeError
            If the provided `version` and `min_version` are not (major, minor,
            patchlevel) tuples of integers.
        """
        for name, ver_spec in {"version": version, "min_version": min_version}.items():
            if (
                not isinstance(ver_spec, tuple)
                or len(ver_spec) != 3
                or not all(isinstance(part, int) for part in ver_spec)
            ):
                raise TypeError(
                    f"The '{name}={ver_spec}' argument must be "
                    "a (major, minor, patchlevel) tuple of integers"
                )
        min_ver_str = f"{min_version[0]}.{min_version[1]}.{min_version[2]}"
        if version >= min_version:
            ver_check = f">= {min_ver_str}, {self._checkmark}"
        else:
            ver_check = f"< {min_ver_str}, {self._nocheckmark}"

        return ver_check


class _NoSubcommand(CotainrSubcommand):
    """A subcommand that simply prints the `parser` help message and exits."""

    def __init__(self, *, parser):
        self.parser = parser

    def execute(self):
        self.parser.print_help()
        sys.exit(0)


class CotainrCLI:
    """
    Build Apptainer/Singularity containers for HPC systems in user space.

    The main cotainr CLI command.

    Parameters
    ----------
    args : list of str, optional
        The input arguments to the CLI (the default is `None`, which implies
        that the input arguments are taken from `sys.argv`).

    Attributes
    ----------
    args : :class:`types.SimpleNamespace`
        The namespace holding the converted arguments parsed to the CLI.
    subcommand : :class:`CotainrSubcommand`
        The subcommand to run.
    """

    _subcommands = [Build, Info]

    def __init__(self, *, args=None):
        """Construct a command line interface for the container builder."""
        # Sanity check subcommands
        for sub_cmd in self._subcommands:
            if not issubclass(sub_cmd, CotainrSubcommand):
                raise TypeError(f"{sub_cmd=} must be a cotainr.cli.CotainrSubcommand")

        # Setup main command parser
        builder_cli_doc_summary = self.__doc__.strip().splitlines()[0]
        parser = argparse.ArgumentParser(
            prog="cotainr", description=builder_cli_doc_summary
        )
        parser.add_argument(
            "--version", action="version", version=f"%(prog)s {_cotainr_version}"
        )

        # Add subcommands parsers
        if self._subcommands:
            subparsers = parser.add_subparsers(title="subcommands")
            for subcommand_cls in self._subcommands:
                subcommand_doc_summary = (
                    subcommand_cls.__doc__.strip().splitlines()[0]
                    if subcommand_cls.__doc__
                    else ""
                )
                sub_parser = subparsers.add_parser(
                    name=subcommand_cls.__name__.lower(),
                    help=subcommand_doc_summary,
                    description=subcommand_doc_summary,
                )
                subcommand_cls.add_arguments(parser=sub_parser)
                sub_parser.set_defaults(subcommand_cls=subcommand_cls)

        # Parse args
        self.args = parser.parse_args(args=args)
        subcommand_args = {
            key: val for key, val in vars(self.args).items() if key != "subcommand_cls"
        }

        # Setup subcommand
        try:
            self.subcommand = self.args.subcommand_cls(**subcommand_args)
        except AttributeError:
            # Print help and exit if no subcommand was given
            self.subcommand = _NoSubcommand(parser=parser)

        # Setup CLI logging
        self._setup_cotainr_cli_logging(
            log_settings=getattr(
                self.subcommand,
                "log_settings",
                tracing.LogSettings(),
            )
        )

    def _setup_cotainr_cli_logging(self, *, log_settings):
        """
        Set up logging for the cotainr main CLI.

        Setting up the logging for the cotainr main CLI includes:
        - Defining log levels based on CLI verbosity arguments.
        - Defining log message formats based on CLI verbosity arguments.
        - Creating log handlers for console output.
        - Creating log handlers for log file output, if requested.
        - Setting up colored console output, if requested.
        - Defining the cotainr "root logger".

        Parameters
        ----------
        log_settings : :class:`~cotainr.tracing.LogSettings`
            The log settings to use for the cotainr main CLI logging.
        """

        class OnlyDebugInfoLevelFilter(logging.Filter):
            """A simple logging filter that removes records >=INFO."""

            def filter(self, record):
                return record.levelno <= logging.INFO

        # Setup cotainr CLI log level and format
        if log_settings.verbosity >= 2:
            cotainr_log_level = logging.DEBUG
            cotainr_stdout_fmt = (
                "%(asctime)s - %(name)s::%(funcName)s::%(lineno)d::"
                "Cotainr:-:%(levelname)s: %(message)s"
            )
            cotainr_stderr_fmt = cotainr_stdout_fmt
        else:
            if log_settings.verbosity <= -1:
                cotainr_log_level = logging.CRITICAL
            else:
                cotainr_log_level = logging.INFO
            cotainr_stdout_fmt = "Cotainr:-: %(message)s"
            cotainr_stderr_fmt = "Cotainr:-:%(levelname)s: %(message)s"

        # Setup cotainr log handlers
        cotainr_stdout_handlers = [logging.StreamHandler(stream=sys.stdout)]
        cotainr_stderr_handlers = [logging.StreamHandler(stream=sys.stderr)]
        if log_settings.log_file_path is not None:
            cotainr_stdout_handlers.append(
                logging.FileHandler(
                    log_settings.log_file_path.with_suffix(
                        log_settings.log_file_path.suffix + ".out"
                    )
                )
            )
            cotainr_stderr_handlers.append(
                logging.FileHandler(
                    log_settings.log_file_path.with_suffix(
                        log_settings.log_file_path.suffix + ".err"
                    )
                )
            )

        for stdout_handler in cotainr_stdout_handlers:
            stdout_handler.setLevel(cotainr_log_level)
            stdout_handler.setFormatter(logging.Formatter(cotainr_stdout_fmt))
            stdout_handler.addFilter(
                # Avoid also emitting WARNINGs and above on stdout
                OnlyDebugInfoLevelFilter()
            )

        for stderr_handler in cotainr_stderr_handlers:
            stderr_handler.setLevel(logging.WARNING)
            stderr_handler.setFormatter(logging.Formatter(cotainr_stderr_fmt))

        if not log_settings.no_color:
            # Replace console formatters with one that colors the output
            cotainr_stdout_handlers[0].setFormatter(
                tracing.ColoredOutputFormatter(cotainr_stdout_fmt)
            )
            cotainr_stderr_handlers[0].setFormatter(
                tracing.ColoredOutputFormatter(cotainr_stderr_fmt)
            )

        # Define cotainr root logger
        root_logger = logging.getLogger("cotainr")
        root_logger.setLevel(cotainr_log_level)
        for handler in cotainr_stdout_handlers + cotainr_stderr_handlers:
            root_logger.addHandler(handler)


def main(*args, **kwargs):
    """Construct the main CLI entrypoint."""
    # Create CotainrCLI to parse command line args and run the specified
    # subcommand
    cli = CotainrCLI()
    cli.subcommand.execute()


def _extract_help_from_docstring(*, arg, docstring):
    """
    Extract the description of `arg` in `string`.

    Parameters
    ----------
    arg : str
        The name of the argument.
    docstring : str
        The numpydoc docstring in which `arg` is documented.

    Returns
    -------
    arg_description : str
        The argument description formatted in a similar way to the default
        arguments provided by an `argparse.ArgumentParser`, i.e. a single line,
        no leading white space, first letter lower case, and no trailing
        period.

    Raises
    ------
    KeyError
        If the docstring does not include `arg`.
    """
    arg_found = False
    arg_desc = []
    for line in docstring.splitlines():
        if arg_found:
            # new command or empty line = end of file.
            # As of Python 3.13 the trailing whitespace is cut and
            # the end of file is not recognized anymore.
            # Therefore, the return is done outside the loop.
            # MARK_PYTHON_VERSION: Revisit once Python 3.12 is deprecated.
            if " : " in line or line.strip() == "":
                break
            else:
                # Extract line as part of arg description
                arg_desc.extend([line.strip(), " "])
        elif f"{arg} : " in line:
            # We found the requested arg in the docstring
            arg_found = True

    if arg_found:
        arg_description = "".join(arg_desc).strip().rstrip(".")
        arg_description = arg_description[0].lower() + arg_description[1:]
        return arg_description
    else:
        # We didn't find the arg in the docstring
        raise KeyError(f"The docstring does not include {arg=}")



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import pathlib

from cotainr.tracing import LogSettings


class TestConstructor:
    def test_correct_types_and_values(self):
        log_settings = LogSettings(
            verbosity=1, log_file_path="/some/path_6021", no_color=True
        )
        assert isinstance(log_settings.verbosity, int)
        assert log_settings.verbosity == 1
        assert isinstance(log_settings.log_file_path, pathlib.Path)
        assert log_settings.log_file_path.parent.as_posix() == "/some"
        assert log_settings.log_file_path.name == "path_6021"
        assert isinstance(log_settings.no_color, bool)
        assert log_settings.no_color

    def test_default_values(self):
        log_settings = LogSettings()
        assert log_settings.verbosity == 0
        assert log_settings.log_file_path is None
        assert not log_settings.no_color



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import pytest

import cotainr.cli


@pytest.fixture
def patch_disable_main(monkeypatch):
    """
    Disable main.

    Replace the `main` cli function with a function that prints and returns a
    message about the execution of the main function that would have run.
    """

    def mock_main(*args, **kwargs):
        msg = "PATCH: Ran cotainr.cli main function"
        print(msg)
        return msg

    monkeypatch.setattr(cotainr.cli, "main", mock_main)


@pytest.fixture
def patch_disable_cotainrcli_init(monkeypatch):
    """
    Make the construction of the CotainrCLI a no-op.
    """

    def mock_init(self, *, args=None):
        pass

    monkeypatch.setattr(cotainr.cli.CotainrCLI, "__init__", mock_init)


@pytest.fixture
def patch_disables_cotainrcli_setup_cotainr_cli_logging(monkeypatch):
    """
    Disable the setup of the logging machinery for the cotainr CLI.

    Replaces the `CotainrCLI._setup_cotainr_cli_logging` method with a method
    that just prints and returns the `tracing.LogSettings` object provided to
    it as an argument.
    """

    def mock_setup_cotainr_cli_logging(self, *, log_settings):
        print(log_settings)
        return log_settings

    monkeypatch.setattr(
        cotainr.cli.CotainrCLI,
        "_setup_cotainr_cli_logging",
        mock_setup_cotainr_cli_logging,
    )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import itertools
import logging
import re

import pytest

from cotainr.cli import CotainrCLI
from cotainr.tracing import LogSettings

from .data import (
    data_cotainr_critical_color_log_messages,
    data_cotainr_debug_color_log_messages,
    data_cotainr_info_color_log_messages,
    data_cotainr_info_no_color_log_messages,
)
from .patches import (
    patch_disable_cotainrcli_init,
    patch_disables_cotainrcli_setup_cotainr_cli_logging,
)
from .stubs import StubInvalidSubcommand, StubLogSettingsSubcommand, StubValidSubcommand


class TestConstructor:
    def test_adding_subcommand(self, argparse_options_line, capsys, monkeypatch):
        monkeypatch.setattr(CotainrCLI, "_subcommands", [StubValidSubcommand])
        with pytest.raises(SystemExit):
            CotainrCLI(args=["stubvalidsubcommand", "-h"])
        stdout = capsys.readouterr().out
        assert stdout == (
            "usage: cotainr stubvalidsubcommand [-h] [--kw-arg KW_ARG] pos_arg\n\n"
            "positional arguments:\n"
            "  pos_arg\n\n"
            f"{argparse_options_line}"
            "  -h, --help       show this help message and exit\n"
            "  --kw-arg KW_ARG\n"
        )

    @pytest.mark.parametrize(
        "subcommands",
        (
            [StubInvalidSubcommand],
            [StubInvalidSubcommand] + CotainrCLI._subcommands,
            CotainrCLI._subcommands + [StubInvalidSubcommand],
        ),
    )
    def test_invalid_subcommands(self, subcommands, monkeypatch):
        monkeypatch.setattr(CotainrCLI, "_subcommands", subcommands)
        with pytest.raises(TypeError, match="must be a cotainr.cli.CotainrSubcommand"):
            CotainrCLI(args=[])

    def test_no_subcommands(self, capsys, monkeypatch):
        monkeypatch.setattr(CotainrCLI, "_subcommands", [])
        with pytest.raises(SystemExit):
            CotainrCLI(args=[]).subcommand.execute()
        stdout = capsys.readouterr().out
        assert stdout.startswith("usage: cotainr [-h] [--version]\n")
        assert "subcommands:" not in stdout

    def test_setup_custom_cli_log_settings_logger(
        self, patch_disables_cotainrcli_setup_cotainr_cli_logging, capsys, monkeypatch
    ):
        monkeypatch.setattr(CotainrCLI, "_subcommands", [StubLogSettingsSubcommand])
        CotainrCLI(
            args=[
                "stublogsettingssubcommand",
                "--verbosity=-1",
                "--log-file-path=/some/path_6021",
                "--no-color",
            ]
        )
        stdout = capsys.readouterr().out.rstrip("\n")
        assert stdout == (
            "LogSettings("
            "verbosity=-1, log_file_path=PosixPath('/some/path_6021'), no_color=True)"
        )

    def test_setup_default_cli_logger(
        self, patch_disables_cotainrcli_setup_cotainr_cli_logging, capsys, monkeypatch
    ):
        monkeypatch.setattr(CotainrCLI, "_subcommands", [StubValidSubcommand])
        CotainrCLI(args=["stubvalidsubcommand", "some_pos_arg_6021"])

        # Check that the default LogSettings are used if no one are specified
        stdout = capsys.readouterr().out.rstrip("\n")
        assert stdout == str(LogSettings())

    @pytest.mark.parametrize(
        "arg, kwarg",
        [
            ("", ""),
            ("another_pos_arg", ""),
            ("", "some_kwarg"),
            ("some_pos_arg", 6021),
        ],
    )
    def test_subcommand_arg_parsing(self, arg, kwarg, capsys, monkeypatch):
        monkeypatch.setattr(CotainrCLI, "_subcommands", [StubValidSubcommand])
        cli = CotainrCLI(args=["stubvalidsubcommand", f"{arg}", f"--kw-arg={kwarg}"])
        cli.subcommand.execute()
        stdout = capsys.readouterr().out
        assert stdout == f"Executed: 'stubvalidsubcommand {arg} --kw-arg={kwarg}'"


class TestHelpMessage:
    cotainr_main_help_msg = (
        # Capsys apparently assumes an 80 char terminal (?) - thus extra '\n'
        "usage: cotainr [-h] [--version] {{build,info}} ...\n\n"
        "Build Apptainer/Singularity containers for HPC systems in user space.\n\n"
        "{argparse_options_line}"
        "  -h, --help    show this help message and exit\n"
        "  --version     show program's version number and exit\n\n"
        "subcommands:\n  {{build,info}}\n"
        "    build       Build a container.\n"
        "    info        Obtain info about the state of all required dependencies for\n"
        "                building a container.\n"
    )

    def test_main_help(self, argparse_options_line, capsys):
        with pytest.raises(SystemExit):
            CotainrCLI(args=["--help"])
        stdout = capsys.readouterr().out
        assert stdout == self.cotainr_main_help_msg.format(
            argparse_options_line=argparse_options_line
        )

    def test_missing_subcommand(self, argparse_options_line, capsys):
        with pytest.raises(SystemExit):
            CotainrCLI(args=[]).subcommand.execute()
        stdout = capsys.readouterr().out
        assert stdout == self.cotainr_main_help_msg.format(
            argparse_options_line=argparse_options_line
        )


class TestVersionMessage:
    def test_main_version(self, capsys):
        from cotainr import __version__ as _cotainr_version

        with pytest.raises(SystemExit):
            CotainrCLI(args=["--version"])
        stdout = capsys.readouterr().out
        assert stdout == f"cotainr {_cotainr_version}\n"


class Test_SetupCotainrCLILogging:
    @pytest.mark.parametrize("verbosity", [-1, -2, -3, -5, -1000])
    def test_cotainr_critical_logging(
        self,
        verbosity,
        capsys,
        data_cotainr_critical_color_log_messages,
        patch_disable_cotainrcli_init,
    ):
        (
            log_level_msgs,
            stdout_msgs,
            stderr_msgs,
        ) = data_cotainr_critical_color_log_messages

        # Setup the CotainrCLI logger
        CotainrCLI()._setup_cotainr_cli_logging(
            log_settings=LogSettings(
                verbosity=verbosity, log_file_path=None, no_color=False
            )
        )
        assert "cotainr" in logging.Logger.manager.loggerDict

        # Log test messages to cotainr root logger
        cotainr_root_logger = logging.getLogger("cotainr")
        for level, msg in log_level_msgs.items():
            cotainr_root_logger.log(level=level, msg=msg)

        # Check correct log level
        assert cotainr_root_logger.getEffectiveLevel() == logging.CRITICAL

        # Check correct handles
        # Pytest manipulates the handler streams to capture the logging output
        assert len(cotainr_root_logger.handlers) == 2
        for handler in cotainr_root_logger.handlers:
            assert isinstance(handler, logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        assert stdout.rstrip("\n").split("\n") == stdout_msgs
        assert stderr.rstrip("\n").split("\n") == stderr_msgs

    @pytest.mark.parametrize("verbosity", [2, 3, 5, 1000])
    def test_cotainr_debug_logging(
        self,
        verbosity,
        capsys,
        data_cotainr_debug_color_log_messages,
        patch_disable_cotainrcli_init,
    ):
        (
            log_level_msgs,
            expected_stdout_msgs,
            expected_stderr_msgs,
        ) = data_cotainr_debug_color_log_messages

        # Setup the CotainrCLI logger
        CotainrCLI()._setup_cotainr_cli_logging(
            log_settings=LogSettings(
                verbosity=verbosity, log_file_path=None, no_color=False
            )
        )
        assert "cotainr" in logging.Logger.manager.loggerDict

        # Log test messages to cotainr root logger
        cotainr_root_logger = logging.getLogger("cotainr")
        for level, msg in log_level_msgs.items():
            cotainr_root_logger.log(level=level, msg=msg)

        # Check correct log level
        assert cotainr_root_logger.getEffectiveLevel() == logging.DEBUG

        # Check correct handles
        # Pytest manipulates the handler streams to capture the logging output
        assert len(cotainr_root_logger.handlers) == 2
        for handler in cotainr_root_logger.handlers:
            assert isinstance(handler, logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        actual_stdout_msgs = stdout.rstrip("\n").split("\n")
        actual_stderr_msgs = stderr.rstrip("\n").split("\n")
        assert len(actual_stdout_msgs) == len(expected_stdout_msgs)
        assert len(actual_stderr_msgs) == len(expected_stderr_msgs)
        debug_msg_start_re = re.compile(
            # Verbose debug msg prefix along lines of:
            # 2023-09-21 10:24:06,191 - cotainr::test_cotainr_debug_logging::163::
            r"^\d{4}(-\d{2}){2} (\d{2}:){2}\d{2},\d{3} - cotainr::test_cotainr_debug_logging::\d+::"
        )
        for actual_msg, expected_msg in itertools.chain(
            zip(actual_stdout_msgs, expected_stdout_msgs),
            zip(actual_stderr_msgs, expected_stderr_msgs),
        ):
            assert debug_msg_start_re.match(actual_msg)
            assert actual_msg.endswith(expected_msg)

    @pytest.mark.parametrize("verbosity", [0, 1])
    def test_cotainr_info_logging(
        self,
        verbosity,
        capsys,
        data_cotainr_info_color_log_messages,
        patch_disable_cotainrcli_init,
    ):
        log_level_msgs, stdout_msgs, stderr_msgs = data_cotainr_info_color_log_messages

        # Setup the CotainrCLI logger
        CotainrCLI()._setup_cotainr_cli_logging(
            log_settings=LogSettings(
                verbosity=verbosity, log_file_path=None, no_color=False
            )
        )
        assert "cotainr" in logging.Logger.manager.loggerDict

        # Log test messages to cotainr root logger
        cotainr_root_logger = logging.getLogger("cotainr")
        for level, msg in log_level_msgs.items():
            cotainr_root_logger.log(level=level, msg=msg)

        # Check correct log level
        assert cotainr_root_logger.getEffectiveLevel() == logging.INFO

        # Check correct handles
        # Pytest manipulates the handler streams to capture the logging output
        assert len(cotainr_root_logger.handlers) == 2
        for handler in cotainr_root_logger.handlers:
            assert isinstance(handler, logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        assert stdout.rstrip("\n").split("\n") == stdout_msgs
        assert stderr.rstrip("\n").split("\n") == stderr_msgs

    @pytest.mark.parametrize("no_color", [True, False])
    def test_log_to_file(
        self,
        no_color,
        tmp_path,
        data_cotainr_info_no_color_log_messages,
        patch_disable_cotainrcli_init,
    ):
        (
            log_level_msgs,
            stdout_msgs,
            stderr_msgs,
        ) = data_cotainr_info_no_color_log_messages

        # Setup the CotainrCLI logger
        log_file_path = tmp_path / "cotainr_log"
        CotainrCLI()._setup_cotainr_cli_logging(
            log_settings=LogSettings(
                verbosity=0, log_file_path=log_file_path, no_color=no_color
            )
        )
        assert "cotainr" in logging.Logger.manager.loggerDict

        # Log test messages to cotainr root logger
        cotainr_root_logger = logging.getLogger("cotainr")
        for level, msg in log_level_msgs.items():
            cotainr_root_logger.log(level=level, msg=msg)

        # Check correct log level
        assert cotainr_root_logger.getEffectiveLevel() == logging.INFO

        # Check correct handles
        # Pytest manipulates the handler streams to capture the logging output
        assert len(cotainr_root_logger.handlers) == 4
        for handler in cotainr_root_logger.handlers[::2]:
            assert isinstance(handler, logging.StreamHandler)
        for handler in cotainr_root_logger.handlers[1::2]:
            assert isinstance(handler, logging.FileHandler)

        # Check correct logging, incl. message format, coloring, log level to file
        assert (
            log_file_path.with_suffix(".out").read_text().rstrip("\n").split("\n")
            == stdout_msgs
        )
        assert (
            log_file_path.with_suffix(".err").read_text().rstrip("\n").split("\n")
            == stderr_msgs
        )

    def test_no_color_on_console(
        self,
        capsys,
        data_cotainr_info_no_color_log_messages,
        patch_disable_cotainrcli_init,
    ):
        (
            log_level_msgs,
            stdout_msgs,
            stderr_msgs,
        ) = data_cotainr_info_no_color_log_messages

        # Setup the CotainrCLI logger
        CotainrCLI()._setup_cotainr_cli_logging(
            log_settings=LogSettings(verbosity=0, log_file_path=None, no_color=True)
        )
        assert "cotainr" in logging.Logger.manager.loggerDict

        # Log test messages to cotainr root logger
        cotainr_root_logger = logging.getLogger("cotainr")
        for level, msg in log_level_msgs.items():
            cotainr_root_logger.log(level=level, msg=msg)

        # Check correct log level
        assert cotainr_root_logger.getEffectiveLevel() == logging.INFO

        # Check correct handles
        # Pytest manipulates the handler streams to capture the logging output
        assert len(cotainr_root_logger.handlers) == 2
        for handler in cotainr_root_logger.handlers:
            assert isinstance(handler, logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        assert stdout.rstrip("\n").split("\n") == stdout_msgs
        assert stderr.rstrip("\n").split("\n") == stderr_msgs



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import argparse
import re
import subprocess

import pytest

from cotainr.cli import CotainrCLI, Info

from ..util.patches import patch_empty_system, patch_system_with_actual_file


class TestExecute:
    def test_report_structure(self, capsys):
        Info().execute()
        stdout = capsys.readouterr().out

        # Dependency report section
        assert stdout.startswith(
            "Dependency report\n"
            "-------------------------------------------------------------------------------\n"
        )
        # One line with python version check
        assert (
            sum(line.startswith("    - Running python ") for line in stdout.split("\n"))
            == 1
        )
        # One line with singularity / apptainer version check
        assert (
            sum(
                any(provider in line for provider in ["singularity-ce", "apptainer"])
                for line in stdout.split("\n")
            )
            == 1
        )

        # Systems info section
        assert (
            "\n"
            "System info\n"
            "-------------------------------------------------------------------------------\n"
        ) in stdout

    def test_with_systems(self, capsys, patch_system_with_actual_file):
        Info().execute()
        stdout = capsys.readouterr().out
        assert stdout.endswith(
            "System info\n"
            "-------------------------------------------------------------------------------\n"
            "Available system configurations:\n"
            "    - some_system_6021\n"
            "    - another_system_6021\n"
        )

    def test_without_systems(self, capsys, patch_empty_system):
        Info().execute()
        stdout = capsys.readouterr().out
        assert stdout.endswith(
            "System info\n"
            "-------------------------------------------------------------------------------\n"
            "No system configurations available\n"
        )


class TestAddArguments:
    def test_no_arguments_added(self):
        parser = argparse.ArgumentParser()
        Info.add_arguments(parser=parser)
        args = parser.parse_args(args=[])
        assert not vars(args)


class TestHelpMessage:
    def test_CLI_subcommand_help_message(self, argparse_options_line, capsys):
        with pytest.raises(SystemExit):
            CotainrCLI(args=["info", "--help"])
        stdout = capsys.readouterr().out
        assert stdout == (
            # Capsys apparently assumes an 80 char terminal (?) - thus extra '\n'
            "usage: cotainr info [-h]\n\n"
            "Obtain info about the state of all required dependencies for building a\n"
            "container.\n\n"
            f"{argparse_options_line}"
            "  -h, --help  show this help message and exit\n"
        )


class Test_check_python_dependency:
    def test_formatting(self):
        # Check for formatting like "Running python 3.11.0 >= 3.9.0, OK"
        # MARK_PYTHON_VERSION: Update this comment to reflect the minimum supported Python version.
        assert re.match(
            (
                r"^Running python \d+\.\d+\.\d+ (>=)|(<) \d+\.\d+\.\d+, "
                r"(\\x1b\[38;5;1mOK\\x1b\[0m)|(\\x1b\[38;5;1mERROR\\x1b\[0m)$"
            ),
            Info()._check_python_dependency(),
        )


class Test_check_singularity_dependency:
    def test_found_apptainer(self, monkeypatch):
        # MARK_APPTAINER_VERSION: Update this to monkeypatch the minimum supported Apptainer version.
        monkeypatch.setattr(
            subprocess,
            "check_output",
            lambda *args, **kwargs: "apptainer version 1.3.4",
        )
        assert (
            Info()._check_singularity_dependency()
            == "Found apptainer 1.3.4 >= 1.3.4, \x1b[38;5;2mOK\x1b[0m"
        )

    def test_found_singularity_ce(self, monkeypatch):
        # MARK_APPTAINER_VERSION: Update this to monkeypatch the minimum supported SingularityCE version.
        monkeypatch.setattr(
            subprocess,
            "check_output",
            lambda *args, **kwargs: "singularity-ce version 3.11.4-1",
        )
        assert (
            Info()._check_singularity_dependency()
            == "Found singularity-ce 3.11.4-1 >= 3.9.2, \x1b[38;5;2mOK\x1b[0m"
        )

    def test_found_unknown(self, monkeypatch):
        monkeypatch.setattr(
            subprocess,
            "check_output",
            lambda *args, **kwargs: "container_tool_6021 version 60.2.1",
        )
        assert (
            Info()._check_singularity_dependency()
            == "Found unknown singularity provider: container_tool_6021 60.2.1"
        )

    def test_not_found(self, monkeypatch):
        def not_found(*args, **kwargs):
            raise FileNotFoundError()

        monkeypatch.setattr(
            subprocess,
            "check_output",
            not_found,
        )
        assert (
            Info()._check_singularity_dependency()
            == "apptainer/singularity not found, \x1b[38;5;1mERROR\x1b[0m"
        )


class Test_check_version:
    @pytest.mark.parametrize(
        "version,min_version,cmp_result",
        [
            ((3, 9, 0), (3, 9, 0), ">= 3.9.0, \x1b[38;5;2mOK\x1b[0m"),
            ((0, 1, 12), (0, 0, 1), ">= 0.0.1, \x1b[38;5;2mOK\x1b[0m"),
            ((3, 7, 14), (3, 9, 0), "< 3.9.0, \x1b[38;5;1mERROR\x1b[0m"),
            ((1, 0, 0), (3, 9, 0), "< 3.9.0, \x1b[38;5;1mERROR\x1b[0m"),
        ],
    )
    def test_version_comparison(self, version, min_version, cmp_result):
        assert (
            Info()._check_version(version=version, min_version=min_version)
            == cmp_result
        )

    def test_wrong_input_len(self):
        with pytest.raises(TypeError) as exc_info:
            Info()._check_version(version=(1, 0, 0, 0), min_version=(1, 0, 0))
        assert str(exc_info.value) == (
            "The 'version=(1, 0, 0, 0)' argument must be "
            "a (major, minor, patchlevel) tuple of integers"
        )

    def test_wrong_input_type(self):
        with pytest.raises(TypeError) as exc_info:
            Info()._check_version(version=(1, 0, 0), min_version="1.0.0")
        assert str(exc_info.value) == (
            "The 'min_version=1.0.0' argument must be "
            "a (major, minor, patchlevel) tuple of integers"
        )

    def test_wrong_part_type(self):
        with pytest.raises(TypeError) as exc_info:
            Info()._check_version(version=("1", "0", "0"), min_version=(1, 0, 0))
        assert str(exc_info.value) == (
            "The 'version=('1', '0', '0')' argument must be "
            "a (major, minor, patchlevel) tuple of integers"
        )
